from datetime import datetime

from pymongo import MongoClient
from pymongo.collection import Collection

# 处理不同优先级
IMMEDIATE = 1
NORMAL = 2
LOW = 3
IDLE = 4


def connect(dbName, collectionName, host="0.0.0.0", port=27017, ) -> Collection:
    client = MongoClient(host=host, port=port)
    db = client[dbName]
    return db[collectionName]


def log_message(message, level, log_db_name="log", log_collection_name="logs", db_host="0.0.0.0", db_port=27017):
    log_collection = connect(log_db_name, log_collection_name, db_host, db_port)
    log_entry = {
        'level': level,
        'message': message,
        'timestamp': datetime.now()
    }
    # 插入日志记录到集合
    log_collection.insert_one(log_entry)



from copy import deepcopy
import numpy as np
from ase import Atoms
from ase.build import make_supercell
from ase.io import read
from ase.data import covalent_radii, chemical_symbols


class VdwBulk:
    """
    A class used to determine which is a Van der Waals interaction of bulk material
    """

    def __init__(self, atoms, connectivity=(0.6, 1.3)):
        self.adj_list = {}
        self.atoms = atoms
        self.clusters = None
        self.connectivity = connectivity
        self._adj_list()

    @staticmethod
    def __get_covalent_radii(symbol):
        '''
        根据元素符号获取元素半径
        :param symbol:symbol,such as 'H'
        '''
        index = chemical_symbols.index(symbol)
        return covalent_radii[index]

    def _adj_list(self):
        """
        A function helps to create a Adjacency list
        """
        symbols = self.atoms.symbols
        positions = self.atoms.positions
        l1 = []
        for i in positions:
            tmp = []
            for j in positions:
                res = np.sqrt(np.sum(np.square(i - j)))
                tmp.append(res)
            l1.append(deepcopy(tmp))
            tmp.clear()
        for i, v in enumerate(l1):
            element = symbols[i]
            tmp = []
            for j, v2 in enumerate(v):
                element2 = symbols[j]
                r1 = self.__get_covalent_radii(element)
                r2 = self.__get_covalent_radii(element2)
                if (r1 + r2) * self.connectivity[0] < v2 < (r1 + r2) * self.connectivity[1]:
                    tmp.append(j)
            self.adj_list[i] = deepcopy(tmp)
            tmp.clear()

    def __search(self, node, adj_list):
        """
        A simplified Depth First Search algorithm
        :param node: the number of node
        :param adj_list: Adjacency list
        :return: all adj_lists which associated with the node
        """
        l1 = adj_list.pop(node)
        for i in l1:
            if i in self.adj_list.keys():
                for i in self.__search(i, adj_list):
                    yield i
            else:
                yield l1
        yield l1

    def __find_cluster_indexes(self, adj_list):
        """
        get all connected graphs
        """
        while self.adj_list.keys():
            node_index = list(adj_list.keys())[0]
            l1 = list(self.__search(node_index, adj_list))
            s1 = {j for i in l1 for j in i}
            yield deepcopy(s1)
            s1.clear()

    def get_cluster(self):
        self.clusters = list(self.__find_cluster_indexes(self.adj_list))
        return self.clusters

    def get_1layer(self):
        """
        Obtain a layer of the bulk material
        :return:one layer of atoms
        """
        if not self.clusters:
            self.get_cluster()
        cluster1 = self.clusters[0]
        atoms_new = Atoms(cell=self.atoms.get_cell())
        for index in cluster1:
            atoms_new.extend(self.atoms[index])
        return atoms_new

    def get_cluster_count(self):
        """
        Get the number of clusters
        """
        if not self.clusters:
            self.get_cluster()
        return len(self.clusters)

    @classmethod
    def dimension_classification(cls, atoms):
        """
        According to the value of count2 / count1,the bulk structure can be classified into five types.
        count2 / count1 == 1: Three dimensional compound
                        == 2: Two dimensional layered compound
                        == 4: One dimensional compound
                        == 8: Zero dimensional compound
                        == other: mixed
        :param atoms:ase.Atoms object
        """
        vdwbulk1 = cls(atoms)
        atoms2x2 = make_supercell(vdwbulk1.atoms, [[2, 0, 0], [0, 2, 0], [0, 0, 2]])
        vdwbulk2 = cls(atoms2x2)
        count1 = vdwbulk1.get_cluster_count()
        count2 = vdwbulk2.get_cluster_count()
        return count2 / count1

    def merge_cluster(self):
        """
        Merge associated clusters.
        """
        cluster = self.get_cluster()
        a, b, c = self.atoms.cell
        print(a, b, c)


if __name__ == '__main__':
    atoms = read('../test/ribbon_test/InBrO_mp-27703_primitive.cif')
    v = VdwBulk(atoms)
    # rt = v.get_cluster()
    print(v.adj_list)



import sys
from argparse import ArgumentParser

from PySide2.QtWidgets import QApplication

from htscf.gui.RibbonBuilder import MainWindow


class CLICommand:
    """Ribbon Builder
    """

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        pass

    @staticmethod
    def run(args, parser):
        app = QApplication(sys.argv)
        win = MainWindow()
        win.show()
        sys.exit(app.exec())



# 处理不用日志类型
from os import getenv

__all__ = ["LOG_DB_CONSTANTS", "LOG_LEVEL_CONSTANTS"]


class LogLevelConstants:
    @property
    def LOG_LEVEL_EMERGENCY(self):
        return 1

    @property
    def LOG_LEVEL_ALERT(self):
        return 2

    @property
    def LOG_LEVEL_CRITICAL(self):
        return 3

    @property
    def LOG_LEVEL_ERROR(self):
        return 4

    @property
    def LOG_LEVEL_WARNING(self):
        return 5

    @property
    def LOG_LEVEL_NOTICE(self):
        return 6

    @property
    def LOG_LEVEL_INFO(self):
        return 7

    @property
    def LOG_LEVEL_DEBUG(self):
        return 8


class LogDbConstants:
    # 日志数据库配置
    @property
    def LOG_DB_NAME(self):
        return getenv("HTSCF_LOG_DB_NAME", "log")

    @property
    def LOG_COLLECTION_NAME(self):
        return getenv("HTSCF_LOG_COLLECTION_NAME", "logs")

    @property
    def LOG_DB_HOST(self):
        return getenv("HTSCF_LOG_DB_HOST", "0.0.0.0")

    @property
    def LOG_DB_PORT(self):
        try:
            return int(getenv("HTSCF_LOG_DB_PORT", 27017))
        except ValueError:
            return 27017

    @property
    def LOG_DB_NAME(self):
        return getenv("HTSCF_LOG_DB_NAME", "log")

    @property
    def LOG_DB_NAME(self):
        return getenv("HTSCF_LOG_DB_NAME", "log")


LOG_DB_CONSTANTS = LogDbConstants()
LOG_LEVEL_CONSTANTS = LogLevelConstants()



from pathlib import Path
from subprocess import Popen, PIPE
import uvicorn
from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel
import uuid

from starlette.responses import FileResponse

app = FastAPI()


class CMD(BaseModel):
    cmd: str


class FilePath(BaseModel):
    path: str


class PYCMD(BaseModel):
    exec_path: str
    py_scripts: str


origins = [
    "http://localhost:8080",
    "https://localhost:3000"
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.post("/sh")
def exec_sh(cmd: CMD) -> bytes:
    popen = Popen(cmd.cmd, stdout=PIPE, stderr=PIPE, shell=True)
    popen.wait()
    error = popen.stderr.read()
    stdout = popen.stdout.read()
    if error:
        raise HTTPException(status_code=500, detail=error.decode("utf-8"))
    return stdout


@app.post("/py")
def exec_py(py_cmd: PYCMD) -> bytes:
    fid = uuid.uuid4()
    fpath = Path(f"/tmp/{fid}")
    fpath.write_text(py_cmd.py_scripts)
    exec_path = Path(py_cmd.exec_path).absolute()
    popen = Popen(f"python {fpath}", stdout=PIPE, stderr=PIPE, shell=True, cwd=exec_path)
    popen.wait()
    error = popen.stderr.read()
    stdout = popen.stdout.read()
    if error:
        raise HTTPException(status_code=500, detail=error.decode("utf-8"))
    return stdout


@app.post("/download")
async def download_file(filepath: FilePath):
    filepath = Path(filepath.path).absolute()
    return FileResponse(path=filepath, filename=filepath.name)


if __name__ == '__main__':
    uvicorn.run("remote_sh:app", port=8892, host="0.0.0.0")



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'UI_BuildRibbon.ui'
##
## Created by: Qt User Interface Compiler version 6.2.4
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
                            QMetaObject, QObject, QPoint, QRect,
                            QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QAction, QBrush, QColor, QConicalGradient,
                           QCursor, QFont, QFontDatabase, QGradient,
                           QIcon, QImage, QKeySequence, QLinearGradient,
                           QPainter, QPalette, QPixmap, QRadialGradient,
                           QTransform)
from PySide2.QtWidgets import (QApplication, QCheckBox, QComboBox, QFormLayout,
                               QGridLayout, QGroupBox, QLabel, QLineEdit,
                               QMainWindow, QMenu, QMenuBar, QPushButton,
                               QSizePolicy, QSpacerItem, QStatusBar, QVBoxLayout,
                               QWidget)


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"Ribbon Builder")
        MainWindow.resize(640, 680)
        MainWindow.setMaximumSize(QSize(640, 680))
        # MainWindow.setStyleSheet(u"\n"
        #                          "QPushButton{\n"
        #                          "    background-color: #4CAF50; /* Green */\n"
        #                          "    border: none;\n"
        #                          "    color: white;\n"
        #                          "\n"
        #                          "\n"
        #                          "	border-radius:4px\n"
        #                          "}\n"
        #                          "QPushButton:hover{\n"
        #                          "	background:rgb(0, 255, 127);\n"
        #                          "	color:black;\n"
        #                          "	cursor:pointer\n"
        #                          "}\n"
        #                          "QLineEdit{\n"
        #                          "	width:40px\n"
        #                          "}\n"
        #                          "")
        self.actionOpen = QAction(MainWindow)
        self.actionOpen.setObjectName(u"actionOpen")
        self.actionExit = QAction(MainWindow)
        self.actionExit.setObjectName(u"actionExit")
        self.actionAbout = QAction(MainWindow)
        self.actionAbout.setObjectName(u"actionAbout")
        self.centralWidget = QWidget(MainWindow)
        self.centralWidget.setObjectName(u"centralWidget")
        self.verticalLayout_2 = QVBoxLayout(self.centralWidget)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.gridLayout = QGridLayout()
        self.gridLayout.setObjectName(u"gridLayout")
        self.rootPath = QLineEdit(self.centralWidget)
        self.rootPath.setObjectName(u"rootPath")
        self.rootPath.setReadOnly(True)

        self.gridLayout.addWidget(self.rootPath, 0, 1, 1, 1)

        self.label_3 = QLabel(self.centralWidget)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout.addWidget(self.label_3, 2, 0, 1, 1)

        self.label = QLabel(self.centralWidget)
        self.label.setObjectName(u"label")

        self.gridLayout.addWidget(self.label, 0, 0, 1, 1)

        self.choiceStructure = QComboBox(self.centralWidget)
        self.choiceStructure.setObjectName(u"choiceStructure")

        self.gridLayout.addWidget(self.choiceStructure, 1, 1, 1, 1)

        self.saveToPath = QPushButton(self.centralWidget)
        self.saveToPath.setObjectName(u"saveToPath")

        self.gridLayout.addWidget(self.saveToPath, 2, 3, 1, 1)

        self.backRootPath = QPushButton(self.centralWidget)
        self.backRootPath.setObjectName(u"backRootPath")

        self.gridLayout.addWidget(self.backRootPath, 0, 3, 1, 1)

        self.nextStructure = QPushButton(self.centralWidget)
        self.nextStructure.setObjectName(u"nextStructure")

        self.gridLayout.addWidget(self.nextStructure, 1, 2, 1, 2)

        self.savedPath = QLineEdit(self.centralWidget)
        self.savedPath.setObjectName(u"savedPath")

        self.gridLayout.addWidget(self.savedPath, 2, 1, 1, 1)

        self.openSavePath = QPushButton(self.centralWidget)
        self.openSavePath.setObjectName(u"openSavePath")

        self.gridLayout.addWidget(self.openSavePath, 2, 2, 1, 1)

        self.label_2 = QLabel(self.centralWidget)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout.addWidget(self.label_2, 1, 0, 1, 1)

        self.openRootPath = QPushButton(self.centralWidget)
        self.openRootPath.setObjectName(u"openRootPath")

        self.gridLayout.addWidget(self.openRootPath, 0, 2, 1, 1)

        self.verticalLayout_2.addLayout(self.gridLayout)

        self.gridLayout_8 = QGridLayout()
        self.gridLayout_8.setObjectName(u"gridLayout_8")
        self.groupBox_2 = QGroupBox(self.centralWidget)
        self.groupBox_2.setObjectName(u"groupBox_2")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.groupBox_2.sizePolicy().hasHeightForWidth())
        self.groupBox_2.setSizePolicy(sizePolicy)
        self.groupBox_2.setMaximumSize(QSize(16777215, 16777215))
        self.gridLayout_3 = QGridLayout(self.groupBox_2)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.formLayout = QFormLayout()
        self.formLayout.setObjectName(u"formLayout")
        self.label_8 = QLabel(self.groupBox_2)
        self.label_8.setObjectName(u"label_8")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.label_8)

        self.vectorA = QComboBox(self.groupBox_2)
        self.vectorA.setObjectName(u"vectorA")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.vectorA)

        self.label_17 = QLabel(self.groupBox_2)
        self.label_17.setObjectName(u"label_17")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.label_17)

        self.vectorB = QComboBox(self.groupBox_2)
        self.vectorB.setObjectName(u"vectorB")

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.vectorB)

        self.label_18 = QLabel(self.groupBox_2)
        self.label_18.setObjectName(u"label_18")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.label_18)

        self.vectorC = QComboBox(self.groupBox_2)
        self.vectorC.setObjectName(u"vectorC")

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.vectorC)

        self.gridLayout_3.addLayout(self.formLayout, 0, 0, 1, 1)

        self.gridLayout_2 = QGridLayout()
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.toggleVectorAB = QPushButton(self.groupBox_2)
        self.toggleVectorAB.setObjectName(u"toggleVectorAB")
        self.toggleVectorAB.setMaximumSize(QSize(130, 16777215))

        self.gridLayout_2.addWidget(self.toggleVectorAB, 0, 0, 1, 2)

        self.setVectorZigzag = QPushButton(self.groupBox_2)
        self.setVectorZigzag.setObjectName(u"setVectorZigzag")
        self.setVectorZigzag.setMaximumSize(QSize(80, 16777215))

        self.gridLayout_2.addWidget(self.setVectorZigzag, 1, 0, 1, 1)

        self.setVectorArmchair = QPushButton(self.groupBox_2)
        self.setVectorArmchair.setObjectName(u"setVectorArmchair")
        self.setVectorArmchair.setMaximumSize(QSize(80, 16777215))

        self.gridLayout_2.addWidget(self.setVectorArmchair, 1, 1, 1, 1)

        self.setVectorHorizontal = QPushButton(self.groupBox_2)
        self.setVectorHorizontal.setObjectName(u"setVectorHorizontal")
        self.setVectorHorizontal.setMaximumSize(QSize(80, 16777215))

        self.gridLayout_2.addWidget(self.setVectorHorizontal, 2, 0, 1, 1)

        self.setVectorVertical = QPushButton(self.groupBox_2)
        self.setVectorVertical.setObjectName(u"setVectorVertical")
        self.setVectorVertical.setMaximumSize(QSize(80, 16777215))

        self.gridLayout_2.addWidget(self.setVectorVertical, 2, 1, 1, 1)

        self.gridLayout_3.addLayout(self.gridLayout_2, 0, 1, 1, 1)

        self.horizontalSpacer_4 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout_3.addItem(self.horizontalSpacer_4, 1, 1, 1, 1)

        self.gridLayout_8.addWidget(self.groupBox_2, 0, 0, 1, 1)

        self.groupBox = QGroupBox(self.centralWidget)
        self.groupBox.setObjectName(u"groupBox")
        sizePolicy.setHeightForWidth(self.groupBox.sizePolicy().hasHeightForWidth())
        self.groupBox.setSizePolicy(sizePolicy)
        self.verticalLayout = QVBoxLayout(self.groupBox)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout_7 = QVBoxLayout()
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.gridLayout_7 = QGridLayout()
        self.gridLayout_7.setObjectName(u"gridLayout_7")
        self.widthControlDelta = QComboBox(self.groupBox)
        self.widthControlDelta.setObjectName(u"widthControlDelta")

        self.gridLayout_7.addWidget(self.widthControlDelta, 1, 1, 1, 1)

        self.widthControlOrigo = QComboBox(self.groupBox)
        self.widthControlOrigo.setObjectName(u"widthControlOrigo")

        self.gridLayout_7.addWidget(self.widthControlOrigo, 2, 1, 1, 1)

        self.setVector1_2 = QPushButton(self.groupBox)
        self.setVector1_2.setObjectName(u"setVector1_2")

        self.gridLayout_7.addWidget(self.setVector1_2, 0, 2, 1, 1)

        self.label_13 = QLabel(self.groupBox)
        self.label_13.setObjectName(u"label_13")

        self.gridLayout_7.addWidget(self.label_13, 0, 0, 1, 1)

        self.setVector2_3 = QPushButton(self.groupBox)
        self.setVector2_3.setObjectName(u"setVector2_3")

        self.gridLayout_7.addWidget(self.setVector2_3, 1, 2, 1, 1)

        self.widthControlBase = QComboBox(self.groupBox)
        self.widthControlBase.setObjectName(u"widthControlBase")

        self.gridLayout_7.addWidget(self.widthControlBase, 0, 1, 1, 1)

        self.ribbonTolerance = QLineEdit(self.groupBox)
        self.ribbonTolerance.setObjectName(u"ribbonTolerance")
        sizePolicy1 = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.ribbonTolerance.sizePolicy().hasHeightForWidth())
        self.ribbonTolerance.setSizePolicy(sizePolicy1)

        self.gridLayout_7.addWidget(self.ribbonTolerance, 3, 1, 1, 1)

        self.label_14 = QLabel(self.groupBox)
        self.label_14.setObjectName(u"label_14")

        self.gridLayout_7.addWidget(self.label_14, 1, 0, 1, 1)

        self.setVector2_2 = QPushButton(self.groupBox)
        self.setVector2_2.setObjectName(u"setVector2_2")

        self.gridLayout_7.addWidget(self.setVector2_2, 0, 3, 1, 1)

        self.label_16 = QLabel(self.groupBox)
        self.label_16.setObjectName(u"label_16")

        self.gridLayout_7.addWidget(self.label_16, 3, 0, 1, 1)

        self.label_15 = QLabel(self.groupBox)
        self.label_15.setObjectName(u"label_15")

        self.gridLayout_7.addWidget(self.label_15, 2, 0, 1, 1)

        self.horizontalSpacer_5 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout_7.addItem(self.horizontalSpacer_5, 3, 2, 1, 1)

        self.horizontalSpacer_6 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout_7.addItem(self.horizontalSpacer_6, 3, 3, 1, 1)

        self.verticalLayout_7.addLayout(self.gridLayout_7)

        self.verticalLayout.addLayout(self.verticalLayout_7)

        self.gridLayout_8.addWidget(self.groupBox, 0, 1, 1, 1)

        self.groupBox_7 = QGroupBox(self.centralWidget)
        self.groupBox_7.setObjectName(u"groupBox_7")
        sizePolicy.setHeightForWidth(self.groupBox_7.sizePolicy().hasHeightForWidth())
        self.groupBox_7.setSizePolicy(sizePolicy)
        self.groupBox_7.setMaximumSize(QSize(16777215, 16777215))
        self.formLayout_4 = QFormLayout(self.groupBox_7)
        self.formLayout_4.setObjectName(u"formLayout_4")
        self.label_12 = QLabel(self.groupBox_7)
        self.label_12.setObjectName(u"label_12")

        self.formLayout_4.setWidget(0, QFormLayout.LabelRole, self.label_12)

        self.repeatA = QLineEdit(self.groupBox_7)
        self.repeatA.setObjectName(u"repeatA")

        self.formLayout_4.setWidget(0, QFormLayout.FieldRole, self.repeatA)

        self.label_19 = QLabel(self.groupBox_7)
        self.label_19.setObjectName(u"label_19")

        self.formLayout_4.setWidget(1, QFormLayout.LabelRole, self.label_19)

        self.repeatB = QLineEdit(self.groupBox_7)
        self.repeatB.setObjectName(u"repeatB")

        self.formLayout_4.setWidget(1, QFormLayout.FieldRole, self.repeatB)

        self.label_20 = QLabel(self.groupBox_7)
        self.label_20.setObjectName(u"label_20")

        self.formLayout_4.setWidget(2, QFormLayout.LabelRole, self.label_20)

        self.repeatC = QLineEdit(self.groupBox_7)
        self.repeatC.setObjectName(u"repeatC")

        self.formLayout_4.setWidget(2, QFormLayout.FieldRole, self.repeatC)

        self.gridLayout_8.addWidget(self.groupBox_7, 1, 0, 1, 1)

        self.groupBox_5 = QGroupBox(self.centralWidget)
        self.groupBox_5.setObjectName(u"groupBox_5")
        sizePolicy.setHeightForWidth(self.groupBox_5.sizePolicy().hasHeightForWidth())
        self.groupBox_5.setSizePolicy(sizePolicy)
        self.formLayout_2 = QFormLayout(self.groupBox_5)
        self.formLayout_2.setObjectName(u"formLayout_2")
        self.label_52 = QLabel(self.groupBox_5)
        self.label_52.setObjectName(u"label_52")

        self.formLayout_2.setWidget(0, QFormLayout.LabelRole, self.label_52)

        self.minRibbonWidth = QLineEdit(self.groupBox_5)
        self.minRibbonWidth.setObjectName(u"minRibbonWidth")

        self.formLayout_2.setWidget(0, QFormLayout.FieldRole, self.minRibbonWidth)

        self.label_50 = QLabel(self.groupBox_5)
        self.label_50.setObjectName(u"label_50")

        self.formLayout_2.setWidget(1, QFormLayout.LabelRole, self.label_50)

        self.maxRibbonWidth = QLineEdit(self.groupBox_5)
        self.maxRibbonWidth.setObjectName(u"maxRibbonWidth")

        self.formLayout_2.setWidget(1, QFormLayout.FieldRole, self.maxRibbonWidth)

        self.label_49 = QLabel(self.groupBox_5)
        self.label_49.setObjectName(u"label_49")

        self.formLayout_2.setWidget(2, QFormLayout.LabelRole, self.label_49)

        self.fixedRibbonWidth = QLineEdit(self.groupBox_5)
        self.fixedRibbonWidth.setObjectName(u"fixedRibbonWidth")

        self.formLayout_2.setWidget(2, QFormLayout.FieldRole, self.fixedRibbonWidth)

        self.gridLayout_8.addWidget(self.groupBox_5, 1, 1, 1, 1)

        self.groupBox_4 = QGroupBox(self.centralWidget)
        self.groupBox_4.setObjectName(u"groupBox_4")
        sizePolicy.setHeightForWidth(self.groupBox_4.sizePolicy().hasHeightForWidth())
        self.groupBox_4.setSizePolicy(sizePolicy)
        self.groupBox_4.setMaximumSize(QSize(16777215, 16777215))
        self.formLayout_6 = QFormLayout(self.groupBox_4)
        self.formLayout_6.setObjectName(u"formLayout_6")
        self.label_22 = QLabel(self.groupBox_4)
        self.label_22.setObjectName(u"label_22")

        self.formLayout_6.setWidget(0, QFormLayout.LabelRole, self.label_22)

        self.vacuumA = QLineEdit(self.groupBox_4)
        self.vacuumA.setObjectName(u"vacuumA")

        self.formLayout_6.setWidget(0, QFormLayout.FieldRole, self.vacuumA)

        self.label_10 = QLabel(self.groupBox_4)
        self.label_10.setObjectName(u"label_10")

        self.formLayout_6.setWidget(1, QFormLayout.LabelRole, self.label_10)

        self.vacuumB = QLineEdit(self.groupBox_4)
        self.vacuumB.setObjectName(u"vacuumB")

        self.formLayout_6.setWidget(1, QFormLayout.FieldRole, self.vacuumB)

        self.label_11 = QLabel(self.groupBox_4)
        self.label_11.setObjectName(u"label_11")

        self.formLayout_6.setWidget(2, QFormLayout.LabelRole, self.label_11)

        self.vacuumC = QLineEdit(self.groupBox_4)
        self.vacuumC.setObjectName(u"vacuumC")

        self.formLayout_6.setWidget(2, QFormLayout.FieldRole, self.vacuumC)

        self.gridLayout_8.addWidget(self.groupBox_4, 2, 0, 1, 1)

        self.groupBox_6 = QGroupBox(self.centralWidget)
        self.groupBox_6.setObjectName(u"groupBox_6")
        sizePolicy.setHeightForWidth(self.groupBox_6.sizePolicy().hasHeightForWidth())
        self.groupBox_6.setSizePolicy(sizePolicy)
        self.formLayout_7 = QFormLayout(self.groupBox_6)
        self.formLayout_7.setObjectName(u"formLayout_7")
        self.label_7 = QLabel(self.groupBox_6)
        self.label_7.setObjectName(u"label_7")
        sizePolicy2 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.label_7.sizePolicy().hasHeightForWidth())
        self.label_7.setSizePolicy(sizePolicy2)
        self.label_7.setAlignment(Qt.AlignLeading | Qt.AlignLeft | Qt.AlignVCenter)

        self.formLayout_7.setWidget(0, QFormLayout.LabelRole, self.label_7)

        self.formLayout_5 = QFormLayout()
        self.formLayout_5.setObjectName(u"formLayout_5")
        self.formLayout_5.setFormAlignment(Qt.AlignLeading | Qt.AlignLeft | Qt.AlignVCenter)
        self.label_9 = QLabel(self.groupBox_6)
        self.label_9.setObjectName(u"label_9")

        self.formLayout_5.setWidget(0, QFormLayout.LabelRole, self.label_9)

        self.connectivityMin = QLineEdit(self.groupBox_6)
        self.connectivityMin.setObjectName(u"connectivityMin")
        sizePolicy3 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(1)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.connectivityMin.sizePolicy().hasHeightForWidth())
        self.connectivityMin.setSizePolicy(sizePolicy3)

        self.formLayout_5.setWidget(0, QFormLayout.FieldRole, self.connectivityMin)

        self.label_23 = QLabel(self.groupBox_6)
        self.label_23.setObjectName(u"label_23")

        self.formLayout_5.setWidget(1, QFormLayout.LabelRole, self.label_23)

        self.connectivityMax = QLineEdit(self.groupBox_6)
        self.connectivityMax.setObjectName(u"connectivityMax")
        sizePolicy3.setHeightForWidth(self.connectivityMax.sizePolicy().hasHeightForWidth())
        self.connectivityMax.setSizePolicy(sizePolicy3)

        self.formLayout_5.setWidget(1, QFormLayout.FieldRole, self.connectivityMax)

        self.formLayout_7.setLayout(0, QFormLayout.FieldRole, self.formLayout_5)

        self.label_45 = QLabel(self.groupBox_6)
        self.label_45.setObjectName(u"label_45")
        sizePolicy2.setHeightForWidth(self.label_45.sizePolicy().hasHeightForWidth())
        self.label_45.setSizePolicy(sizePolicy2)

        self.formLayout_7.setWidget(1, QFormLayout.LabelRole, self.label_45)

        self.bottom = QLineEdit(self.groupBox_6)
        self.bottom.setObjectName(u"bottom")
        sizePolicy3.setHeightForWidth(self.bottom.sizePolicy().hasHeightForWidth())
        self.bottom.setSizePolicy(sizePolicy3)

        self.formLayout_7.setWidget(1, QFormLayout.FieldRole, self.bottom)

        self.label_21 = QLabel(self.groupBox_6)
        self.label_21.setObjectName(u"label_21")
        sizePolicy2.setHeightForWidth(self.label_21.sizePolicy().hasHeightForWidth())
        self.label_21.setSizePolicy(sizePolicy2)

        self.formLayout_7.setWidget(2, QFormLayout.LabelRole, self.label_21)

        self.top = QLineEdit(self.groupBox_6)
        self.top.setObjectName(u"top")
        sizePolicy3.setHeightForWidth(self.top.sizePolicy().hasHeightForWidth())
        self.top.setSizePolicy(sizePolicy3)

        self.formLayout_7.setWidget(2, QFormLayout.FieldRole, self.top)

        self.gridLayout_8.addWidget(self.groupBox_6, 2, 1, 1, 1)

        self.groupBox_3 = QGroupBox(self.centralWidget)
        self.groupBox_3.setObjectName(u"groupBox_3")
        self.groupBox_3.setMaximumSize(QSize(16777215, 16777215))
        self.formLayout_3 = QFormLayout(self.groupBox_3)
        self.formLayout_3.setObjectName(u"formLayout_3")
        self.label_4 = QLabel(self.groupBox_3)
        self.label_4.setObjectName(u"label_4")

        self.formLayout_3.setWidget(0, QFormLayout.LabelRole, self.label_4)

        self.ifWidthTest = QCheckBox(self.groupBox_3)
        self.ifWidthTest.setObjectName(u"ifWidthTest")

        self.formLayout_3.setWidget(0, QFormLayout.FieldRole, self.ifWidthTest)

        self.label_5 = QLabel(self.groupBox_3)
        self.label_5.setObjectName(u"label_5")

        self.formLayout_3.setWidget(1, QFormLayout.LabelRole, self.label_5)

        self.ifShowTargetStructure = QCheckBox(self.groupBox_3)
        self.ifShowTargetStructure.setObjectName(u"ifShowTargetStructure")

        self.formLayout_3.setWidget(1, QFormLayout.FieldRole, self.ifShowTargetStructure)

        self.label_6 = QLabel(self.groupBox_3)
        self.label_6.setObjectName(u"label_6")

        self.formLayout_3.setWidget(2, QFormLayout.LabelRole, self.label_6)

        self.terminateAtom = QComboBox(self.groupBox_3)
        self.terminateAtom.setObjectName(u"terminateAtom")
        self.terminateAtom.setMaximumSize(QSize(80, 16777215))

        self.formLayout_3.setWidget(2, QFormLayout.FieldRole, self.terminateAtom)

        self.gridLayout_8.addWidget(self.groupBox_3, 3, 0, 1, 1)

        self.verticalLayout_2.addLayout(self.gridLayout_8)

        self.gridLayout_4 = QGridLayout()
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.showStructure = QPushButton(self.centralWidget)
        self.showStructure.setObjectName(u"showStructure")

        self.gridLayout_4.addWidget(self.showStructure, 0, 0, 1, 1)

        self.runProgram = QPushButton(self.centralWidget)
        self.runProgram.setObjectName(u"runProgram")

        self.gridLayout_4.addWidget(self.runProgram, 0, 1, 1, 1)

        self.stopProgram = QPushButton(self.centralWidget)
        self.stopProgram.setObjectName(u"stopProgram")

        self.gridLayout_4.addWidget(self.stopProgram, 0, 2, 1, 1)

        self.verticalLayout_2.addLayout(self.gridLayout_4)

        MainWindow.setCentralWidget(self.centralWidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 640, 26))
        self.menu = QMenu(self.menubar)
        self.menu.setObjectName(u"menu")
        self.menuEdit = QMenu(self.menubar)
        self.menuEdit.setObjectName(u"menuEdit")
        self.menuWindow = QMenu(self.menubar)
        self.menuWindow.setObjectName(u"menuWindow")
        self.menuSettings = QMenu(self.menubar)
        self.menuSettings.setObjectName(u"menuSettings")
        self.menuHelp = QMenu(self.menubar)
        self.menuHelp.setObjectName(u"menuHelp")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.menubar.addAction(self.menu.menuAction())
        self.menubar.addAction(self.menuEdit.menuAction())
        self.menubar.addAction(self.menuWindow.menuAction())
        self.menubar.addAction(self.menuSettings.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        self.menu.addAction(self.actionOpen)
        self.menu.addAction(self.actionExit)
        self.menuHelp.addAction(self.actionAbout)

        self.retranslateUi(MainWindow)
        self.openRootPath.clicked.connect(MainWindow.openFilesSlot)
        self.backRootPath.clicked.connect(MainWindow.openFilesSlot)
        self.openSavePath.clicked.connect(MainWindow.openFilesSlot)
        self.nextStructure.clicked.connect(MainWindow.nextStructureSlot)
        self.rootPath.textChanged.connect(MainWindow.loadStructuresSlot)
        self.choiceStructure.currentTextChanged.connect(MainWindow.parseStructureSlot)
        self.saveToPath.clicked.connect(MainWindow.saveToPathSlot)
        self.setVector1_2.clicked.connect(MainWindow.changeComboBoxSlot)
        self.setVector2_2.clicked.connect(MainWindow.changeComboBoxSlot)
        self.setVector2_3.clicked.connect(MainWindow.changeComboBoxSlot)
        self.toggleVectorAB.clicked.connect(MainWindow.changeComboBoxSlot)
        self.setVectorZigzag.clicked.connect(MainWindow.changeComboBoxSlot)
        self.setVectorArmchair.clicked.connect(MainWindow.changeComboBoxSlot)
        self.setVectorHorizontal.clicked.connect(MainWindow.changeComboBoxSlot)
        self.setVectorVertical.clicked.connect(MainWindow.changeComboBoxSlot)
        self.runProgram.clicked.connect(MainWindow.taskHandlerSlot)
        self.stopProgram.clicked.connect(MainWindow.taskHandlerSlot)
        self.showStructure.clicked.connect(MainWindow.showStructureSlot)
        self.actionOpen.triggered.connect(MainWindow.openFilesSlot)
        self.actionExit.triggered.connect(MainWindow.close)
        self.actionAbout.triggered.connect(MainWindow.getHelp)
        self.nextStructure.clicked.connect(MainWindow.changeShowStructureSlot)

        QMetaObject.connectSlotsByName(MainWindow)

    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"Ribbon Builder", None))
        self.actionOpen.setText(QCoreApplication.translate("MainWindow", u"&Open", None))
        self.actionExit.setText(QCoreApplication.translate("MainWindow", u"Exit", None))
        self.actionAbout.setText(QCoreApplication.translate("MainWindow", u"About", None))
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Save Path:", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Root:", None))
        self.saveToPath.setText(QCoreApplication.translate("MainWindow", u"Save", None))
        self.backRootPath.setText(QCoreApplication.translate("MainWindow", u"Back", None))
        self.nextStructure.setText(QCoreApplication.translate("MainWindow", u"Next", None))
        self.openSavePath.setText(QCoreApplication.translate("MainWindow", u"Open", None))
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"Structure:", None))
        self.openRootPath.setText(QCoreApplication.translate("MainWindow", u"Open", None))
        self.groupBox_2.setTitle(QCoreApplication.translate("MainWindow", u"Vector", None))
        self.label_8.setText(QCoreApplication.translate("MainWindow", u"a", None))
        self.label_17.setText(QCoreApplication.translate("MainWindow", u"b", None))
        self.label_18.setText(QCoreApplication.translate("MainWindow", u"c", None))
        self.toggleVectorAB.setText(QCoreApplication.translate("MainWindow", u"toggleVector A B", None))
        self.setVectorZigzag.setText(QCoreApplication.translate("MainWindow", u"Zigzag", None))
        self.setVectorArmchair.setText(QCoreApplication.translate("MainWindow", u"Armchair", None))
        self.setVectorHorizontal.setText(QCoreApplication.translate("MainWindow", u"Horizontal", None))
        self.setVectorVertical.setText(QCoreApplication.translate("MainWindow", u"Vertical", None))
        self.groupBox.setTitle(QCoreApplication.translate("MainWindow", u"Width Control", None))
        self.setVector1_2.setText(QCoreApplication.translate("MainWindow", u"1-2", None))
        self.label_13.setText(QCoreApplication.translate("MainWindow", u"Base:", None))
        self.setVector2_3.setText(QCoreApplication.translate("MainWindow", u"2-3", None))
        self.label_14.setText(QCoreApplication.translate("MainWindow", u"Delta:", None))
        self.setVector2_2.setText(QCoreApplication.translate("MainWindow", u"2-2", None))
        self.label_16.setText(QCoreApplication.translate("MainWindow", u"Ribbon Tolerance:", None))
        self.label_15.setText(QCoreApplication.translate("MainWindow", u"Origo:", None))
        self.groupBox_7.setTitle(QCoreApplication.translate("MainWindow", u"Repeat", None))
        self.label_12.setText(QCoreApplication.translate("MainWindow", u"a", None))
        self.label_19.setText(QCoreApplication.translate("MainWindow", u"b", None))
        self.label_20.setText(QCoreApplication.translate("MainWindow", u"c", None))
        self.groupBox_5.setTitle(QCoreApplication.translate("MainWindow", u"Ribbon Width", None))
        self.label_52.setText(QCoreApplication.translate("MainWindow", u"Min", None))
        self.label_50.setText(QCoreApplication.translate("MainWindow", u"Max", None))
        self.label_49.setText(QCoreApplication.translate("MainWindow", u"Fixed", None))
        self.groupBox_4.setTitle(QCoreApplication.translate("MainWindow", u"Vacuum", None))
        self.label_22.setText(QCoreApplication.translate("MainWindow", u"A-Axis:", None))
        self.vacuumA.setText(QCoreApplication.translate("MainWindow", u"15", None))
        self.label_10.setText(QCoreApplication.translate("MainWindow", u"B-Axis:", None))
        self.vacuumB.setText(QCoreApplication.translate("MainWindow", u"15", None))
        self.label_11.setText(QCoreApplication.translate("MainWindow", u"C-Axis:", None))
        self.vacuumC.setText(QCoreApplication.translate("MainWindow", u"15", None))
        self.groupBox_6.setTitle("")
        self.label_7.setText(QCoreApplication.translate("MainWindow", u"Connectivity", None))
        self.label_9.setText(QCoreApplication.translate("MainWindow", u"min:", None))
        self.label_23.setText(QCoreApplication.translate("MainWindow", u"max:", None))
        self.label_45.setText(QCoreApplication.translate("MainWindow", u"Bottom:", None))
        self.label_21.setText(QCoreApplication.translate("MainWindow", u"Top:", None))
        self.groupBox_3.setTitle(QCoreApplication.translate("MainWindow", u"Options", None))
        self.label_4.setText(QCoreApplication.translate("MainWindow", u"Width Test", None))
        self.ifWidthTest.setText("")
        self.label_5.setText(QCoreApplication.translate("MainWindow", u"Show Target Structure", None))
        self.ifShowTargetStructure.setText("")
        self.label_6.setText(QCoreApplication.translate("MainWindow", u"Terminate Atom:", None))
        self.showStructure.setText(QCoreApplication.translate("MainWindow", u"Show", None))
        self.runProgram.setText(QCoreApplication.translate("MainWindow", u"Run", None))
        self.stopProgram.setText(QCoreApplication.translate("MainWindow", u"Stop", None))
        self.menu.setTitle(QCoreApplication.translate("MainWindow", u"File", None))
        self.menuEdit.setTitle(QCoreApplication.translate("MainWindow", u"Edit", None))
        self.menuWindow.setTitle(QCoreApplication.translate("MainWindow", u"Window", None))
        self.menuSettings.setTitle(QCoreApplication.translate("MainWindow", u"Settings", None))
        self.menuHelp.setTitle(QCoreApplication.translate("MainWindow", u"Help", None))
    # retranslateUi



import os
import re
import subprocess
import uuid
from functools import wraps
from hashlib import md5
from inspect import getattr_static
from io import StringIO
from os.path import join
from pathlib import Path
from re import Pattern, match
from shutil import copy
from subprocess import Popen, PIPE
from typing import List, Union, Iterator
from uuid import uuid4
from datetime import datetime
from ase import Atoms
from ase.io import read
from ase.io.cif import CIFBlock, parse_block
from ase.io.cif_unicode import format_unicode


def py2cmdline(script, args: list = None):
    if args is None:
        args = []
    args = map(lambda i: str(i), args)
    exs = 'exec(%r)' % re.sub('\r\n|\r', '\\n', script.rstrip())
    return 'python -c "%s"' % exs.replace('"', r'\"').replace("$", r"\$") + " " + " ".join(args)


def copyFile(src_dir, target_dir, file):
    if not os.path.exists(target_dir):
        os.makedirs(target_dir)
    src = join(src_dir, file)
    copy(src, target_dir)


def search(fp, target, exclude=None, recursion=False):
    """
    Search files.
    :param fp: file path.
    :param target: target filename or keywords.
    :param exclude: exclude filename or keywords.
    :param recursion: Whether to recursively search
    :return: a generator
    """
    abs_fp = os.path.abspath(fp)
    items = os.listdir(abs_fp)
    for item in items:
        item_fp = os.path.join(abs_fp, item)
        if os.path.isdir(item_fp) and recursion:
            for tmp in search(item_fp, target, exclude, recursion):
                yield tmp
        elif exclude:
            if target in item and exclude not in item:
                yield item_fp
        elif target in item:
            yield item_fp


def md5_hex(file: Union[str, Path]):
    """计算16进制的md5值
    """
    return md5(Path(file).read_bytes()).hexdigest()


def matchList(li: list, pattern: Union[str, Pattern], return_type="str"):
    """匹配列表中符合条件的第一列
    """
    for i in li:
        if match(pattern, i):
            if return_type == "str":
                return i
            elif return_type == "Match":
                return match(pattern, i)
    return ""


def matchManyList(li: list, patterns: List[Union[str, Pattern]]):
    return [matchList(li, pattern) for pattern in patterns]


def method_register(cls):
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            return func(self, *args, **kwargs)

        if getattr_static(cls, func.__name__, None):
            msg = 'Error method name REPEAT, {} has exist'.format(func.__name__)
            raise NameError(msg)
        else:
            setattr(cls, func.__name__, wrapper)
        return func

    return decorator


def parse_cif_ase(s) -> Iterator[CIFBlock]:
    """Parse a CIF file using ase CIF parser."""

    data = format_unicode(s)
    lines = [e for e in data.split('\n') if len(e) > 0]
    lines = [''] + lines[::-1]  # all lines (reversed)

    while lines:
        line = lines.pop().strip()
        if not line or line.startswith('#'):
            continue

        yield parse_block(lines, line)


def parse_cif_ase_string(s) -> Atoms:
    return list(parse_cif_ase(s))[0].get_atoms()


def cif2poscar(cif_content: str):
    sio = StringIO()
    atoms = parse_cif_ase_string(cif_content)
    atoms.write(sio, format="vasp")
    sio.seek(0)
    lines = sio.readlines()
    symbol = lines[5].split()
    number = lines[6].split()
    positions = lines[8:]
    d = {}
    for s, n in zip(symbol, number):
        n = int(n)
        if s not in d.keys():
            d.update({s: [0, []]})
        d[s][0] = d[s][0] + n
        d[s][1].extend([positions.pop(0) for _ in range(n)])
    _symbol = []
    _number = []
    _positions = []
    for k, v in d.items():
        _symbol.append(k)
        _number.append(str(v[0]))
        _positions.extend(v[1])
    name = "".join([f"{s}{n}" for s, n in zip(_symbol, _number)])
    content = f"{name}\n{''.join(lines[1:5])}{' ' + ' '.join(_symbol)}\n{'  ' + ' '.join(_number)}\n{lines[7]}{''.join(_positions)}"
    return content


def writeScript(rootPath: Path, script) -> Path:
    scriptFileName = f"script-{uuid4()}"  # 生成一个随机脚本名
    scriptPath = rootPath / scriptFileName
    scriptPath.write_text(script)
    return scriptPath.absolute()


def mpiRun(command: str, directory: Path, stdout_file="vasp.out", stderr_file="vasp.error") -> str:
    start_time = datetime.now()
    stdout_path = directory / stdout_file
    stderr_path = directory / stderr_file

    # 生成唯一的文件名后缀
    unique_suffix = uuid.uuid4()

    # 准备命令脚本
    mpi_commands = " && ".join([
        'NP=$(cat "$PBS_NODEFILE" | wc -l)',
        f'cat "$PBS_NODEFILE" | sort | uniq | tee /tmp/nodes.{unique_suffix} | wc -l',
        f'cat "$PBS_NODEFILE" > /tmp/nodefile.{unique_suffix}',
        f'mpirun -genv I_MPI_DEVICE ssm -machinefile /tmp/nodefile.{unique_suffix} -n "$NP" {command}',
        f'rm -rf /tmp/nodefile.{unique_suffix}',
        f'rm -rf /tmp/nodes.{unique_suffix}'
    ])

    with stdout_path.open('w') as stdout_file, stderr_path.open('w') as stderr_file:
        process = Popen(mpi_commands, cwd=directory, stdout=stdout_file, stderr=stderr_file, shell=True, executable='/bin/bash')
        process.wait()

    end_time = datetime.now()
    return (end_time - start_time).__str__()


def is_converge(filepath: Path) -> bool:
    """
    Checks if the calculation has converged based on the presence of a specific
    string in the OUTCAR file.

    Args:
        filepath (Path): The directory path where the OUTCAR file is located.

    Returns:
        bool: True if the calculation has converged, False otherwise.
    """
    outcar = filepath / "OUTCAR"
    try:
        with outcar.open() as fd:
            for line in fd:
                if "reached required accuracy" in line:
                    return True
        return False
    except FileNotFoundError:
        print(f"File {outcar} not found.")
        return False


def read_from_str(s: str, format_="cif") -> Atoms:
    """
    从字符串中读取结构到ase中
    例(从字符串读取vasp文件):
        atoms = read_from_str(
        "S12Al2P4Li2H4
1.0000000000000000
 6.8544000000000000    0.0000000000000000    0.0000000000000000
 0.0000000000000000   31.2483000000000004    0.0000000000000000
 0.0000000000000000    0.0000000000000000   18.0067999999999984
S Al P Li H
12 2 4 2 4
Cartesian
1.1436566399999999 19.2436405889999982  0.0682457720000000
4.5708566400000015 14.0357989110000005  0.0682457720000000
1.4675955839999999 15.8453879639999986  0.0000000000000000
5.3868044159999995 15.8453879639999986  3.0067754639999995",format="vasp")
    """
    with StringIO() as sio:
        sio.write(s)
        sio.seek(0)
        return read(sio, format=format_)


class Pipe:
    def __init__(self, command, current_directory, encoding="utf-8"):
        self.command = command
        self.current_directory = current_directory
        self.encoding = encoding
        self.stdout = ""
        self.stderr = ""
        self.pipe()

    def pipe(self):
        popen = Popen(self.command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True,
                      cwd=self.current_directory)
        stdout, stderr = popen.communicate()
        if stdout:
            self.stdout = stdout.decode(self.encoding)
            return
        if stderr:
            self.stderr = stderr.decode(self.encoding)


def execute_command(command, shell=False):
    with Popen(command, stdout=PIPE, stderr=PIPE, shell=shell, text=True) as process:
        stdout, stderr = process.communicate()
        if process.returncode != 0:
            raise RuntimeError(f"Command failed with exit code {process.returncode}")
        return stdout, stderr


def del_nested_attr(obj, attr):
    """
    尝试删除一个嵌套属性，如果成功删除则返回 True，否则返回 False。

    :param obj: 目标对象或字典
    :param attr: 属性的路径，例如 "step_details.file"
    :return: 布尔值，表示属性是否被成功删除
    """
    # 分割属性路径
    parts = attr.split('.')

    # 遍历到最后一个属性之前的路径
    for part in parts[:-1]:
        if not isinstance(obj, dict) or part not in obj:
            return False
        obj = obj[part]

    # 删除最终的属性
    if isinstance(obj, dict) and parts[-1] in obj:
        del obj[parts[-1]]
        return True
    else:
        return False



#!/usr/bin/env python

import os
import re
import sys
from argparse import ArgumentParser
from optparse import OptionParser
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.ticker import AutoMinorLocator
import numpy
import numpy as np


class CLICommand:
    """能带绘制
    """

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        add = parser.add_argument

        add('-f', '--file',
            action='store', type=str,
            dest='filename', default='OUTCAR',
            help='location of OUTCAR')

        add('--procar',
            action='store', type=str, dest='procar',
            default='PROCAR',
            help='location of the PROCAR')

        add('-z', '--zero',
            action='store', type=float,
            dest='efermi', default=None,
            help='energy reference of the band plot')

        add('--adjust_gap',
            action='store', type=float,
            dest='adjust_gap', default=None,
            help='基于当前gap进行调整')

        add('-o', '--output',
            action='store', type=str, dest='bandimage',
            default='band.png',
            help='output image name, "band.png" by default')

        add('-k', '--kpoints',
            action='store', type=str, dest='kpts',
            default=None,
            help='kpoint path')

        add('-s', '--size', nargs=2,
            action='store', type=float, dest='figsize',
            default=(3.0, 4.0),
            help='figure size of the output plot')

        add('-y', nargs=2,
            action='store', type=float, dest='ylim',
            default=(-3, 3),
            help='energy range of the band plot')

        add('--hline',
            action='append', type=float, dest='hlines',
            default=[],
            help='Add horizontal lines to the figure.')

        add('--vline',
            action='append', type=float, dest='vlines',
            default=[],
            help='Add vertical lines to the figure.')

        add('--save_gnuplot',
            action='store_true', dest='gnuplot',
            default=False,
            help='save output band energies in gnuplot format')

        add('--lw',
            action='store', type=float, dest='linewidth',
            default=1.0,
            help='linewidth of the band plot')

        add('--lc',
            action='store', type=str, dest='linecolors',
            default=None,
            help='line colors of the band plot')

        add('--dpi',
            action='store', type=int, dest='dpi',
            default=360,
            help='resolution of the output image')

        add('--occ',
            action='append', type=str, dest='occ',
            default=[],
            help='orbital contribution of each KS state')

        add('--occL',
            action='store_true', dest='occLC',
            default=False,
            help='use Linecollection or Scatter to show the orbital contribution')

        add('--occLC_cmap',
            action='store', type=str, dest='occLC_cmap',
            default='jet',
            help='colormap of the line collection')

        add('--occLC_lw',
            action='store', type=float, dest='occLC_lw',
            default=2.0,
            help='linewidth of the line collection')

        add('--occLC_cbar_pos',
            action='store', type=str, dest='occLC_cbar_pos',
            default='top',
            help='position of the colorbar')

        add('--occLC_cbar_ticks',
            action='store', type=str, dest='occLC_cbar_ticks',
            default=None,
            help='ticks for the colorbar')

        add('--occLC_cbar_vmin',
            action='store', type=float, dest='occLC_cbar_vmin',
            default=None,
            help='minimum value for the color plot')

        add('--occLC_cbar_vmax',
            action='store', type=float, dest='occLC_cbar_vmax',
            default=None,
            help='maximum value for the color plot')

        add('--occLC_cbar_ticklabels',
            action='store', type=str, dest='occLC_cbar_ticklabels',
            default=None,
            help='tick labels for the colorbar')

        add('--occLC_cbar_size',
            action='store', type=str, dest='occLC_cbar_size',
            default='3%',
            help='size of the colorbar, relative to the axis')

        add('--occLC_cbar_pad',
            action='store', type=float, dest='occLC_cbar_pad',
            default=0.02,
            help='pad between colorbar and axis')

        add('--occM',
            action='append', type=str, dest='occMarker',
            default=[],
            help='the marker used in the plot')

        add('--occMs',
            action='append', type=int, dest='occMarkerSize',
            default=[],
            help='the size of the marker')

        add('--occMc',
            action='append', type=str, dest='occMarkerColor',
            default=[],
            help='the color of the marker')

        add('--spd',
            action='append', type=str, dest='spdProjections',
            default=[],
            help='Spd-projected wavefunction character of each KS orbital.')

        add('--spin', action='store', dest='spin',
            default=None, choices=['x', 'y', 'z'],
            help='show the magnetization mx/y/z constributions to the states. Use this option along with --occ.')

        add('--lsorbit',
            action='store_true', dest='lsorbit',
            help='Spin orbit coupling on, special treament of PROCAR')

        add('-q', '--quiet',
            action='store_true', dest='quiet',
            help='not show the resulting image')

        add('--xlabel',
            action='store', type=str,
            dest='xlabel', default=None,
            help='设置x坐标描述')

        add('--ylabel',
            action='store', type=str,
            dest='ylabel', default=None,
            help='设置y坐标描述')

    @staticmethod
    def run(args, parser):

        if args.occ:

            Nocc = len(args.occ)
            occM = ['o' for ii in range(Nocc)]
            occMc = ['r' for ii in range(Nocc)]
            occMs = [20 for ii in range(Nocc)]
            for ii in range(min(len(args.occMarker), Nocc)):
                occM[ii] = args.occMarker[ii]
            for ii in range(min(len(args.occMarkerSize), Nocc)):
                occMs[ii] = args.occMarkerSize[ii]
            for ii in range(min(len(args.occMarkerColor), Nocc)):
                occMc[ii] = args.occMarkerColor[ii]
            args.occMarker = occM
            args.occMarkerColor = occMc
            args.occMarkerSize = occMs

            whts = []
            iterater = 0
            for occ in args.occ:
                # CCX 2020-01-31 parse input range of atoms
                # if '0' in the index, select all the atoms
                if '0' in occ.split():
                    occAtom = None
                else:
                    occAtom = parseList(occ)

                if args.spdProjections:  # and (Nocc == 1):
                    if len(args.spdProjections) != len(args.occ):
                        print("number of projections does not match number of occupations")
                        sys.exit(0)

                    # set angall to corresponding item of spdProjections
                    angularM = parseSpdProjection(args.spdProjections[iterater])

                    whts.append(WeightFromPro(args.procar, whichAtom=occAtom,
                                              spd=angularM, lsorbit=args.lsorbit,
                                              spin=args.spin))
                else:
                    whts.append(WeightFromPro(args.procar, whichAtom=occAtom,
                                              lsorbit=args.lsorbit,
                                              spin=args.spin))
                iterater += 1

        else:
            whts = None

        kpath, bands, efermi, kpt_bounds = get_bandInfo(args.filename)
        bands = numpy.array([bands[0]])
        if args.efermi is None:
            bands -= efermi
        else:
            bands -= args.efermi
        if args.adjust_gap:
            bands[bands > 0] += args.adjust_gap

        import matplotlib as mpl
        from matplotlib.ticker import AutoMinorLocator

        # Use non-interactive backend in case there is no display
        mpl.use('agg')
        import matplotlib.pyplot as plt

        mpl.rcParams['axes.unicode_minus'] = False

        mpl_default_colors_cycle = [mpl.colors.to_hex(xx) for xx in
                                    mpl.rcParams['axes.prop_cycle'].by_key()['color']]
        if args.linecolors:
            ctmp = [mpl.colors.to_hex(xx) for xx in args.linecolors.split()]
            nspin = bands.shape[0]
            if len(ctmp) <= nspin:
                args.linecolors = ctmp + \
                                  [xx for xx in mpl_default_colors_cycle if xx not in ctmp]
        else:
            args.linecolors = mpl_default_colors_cycle

        bandplot(kpath, bands, efermi, kpt_bounds, args, whts)
        saveband_dat(kpath, bands, args, whts)


def parseList(string):
    def parseRange(rng):
        # print(rng)
        m = re.match(r'(\d+)(?:[-:](\d+))?(?:[-:](\d+))?$', rng)
        if not m:
            raise ValueError(
                """
The index should be assigned with combination of the following ways:
-> 10, a single band with index 10
-> 20:30, or '20-30', a continuous range from 20 to 30, 30 included
-> 30:50:2, or '30-50:2', a continues range from 30 to 50, with step size 2
-> '1 2 3', all the patterns above in one string separated by spaces.
For example: '1 4:6 8:16:2' will be converted to '1 4 5 6 8 10 12 14 16'
"""
            )
        ii = m.group(1)
        jj = m.group(2) or ii
        ss = m.group(3) or 1
        return [x - 1 for x in range(int(ii), int(jj) + 1, int(ss))]

    ret = []
    for rng in string.split():
        ret += parseRange(rng)
    return list(set(ret))


def parseSpdProjection(spd):
    '''
    Parse spdProjections string.  str -> [int]

    # Ionizing
    '''
    spd_dict = {
        's': [0],
        'p': [1, 2, 3],
        'd': [4, 5, 6, 7, 8],
        'f': [9, 10, 11, 12, 13, 14, 15],
        'py': [1],
        'pz': [2],
        'px': [3],
        'dxy': [4],
        'dyz': [5],
        'dz2': [6],
        'dxz': [7],
        'dx2': [8],
        "fy(3x2-y2)": [9],
        "fxyz  ": [10],
        "fyz2  ": [11],
        "fz3   ": [12],
        "fxz2  ": [13],
        "fz(x2-y2)": [14],
        "fx(x2-3y2) ": [15],
    }

    ret = []
    for l in spd.split():
        try:
            assert int(l) <= 15, "Maximum spd index should be <= 15."
            ret += [int(l)]
        except:
            if l.lower() not in spd_dict:
                raise ValueError(
                    "Spd-projected wavefunction character of each KS orbital.\n"
                    "    s orbital: 0\n"
                    "    py, pz, px orbital: 1 2 3\n"
                    "    dxy, dyz, dz2, dxz, dx2 orbital: 4 5 6 7 8 \n"
                    "    fy(3x2-y2), fxyz, fyz2, fz3, fxz2, fz(x2-y2), fx(x2-3y2) orbital: 9 10 11 12 13 14 15\n"
                    "\nFor example, --spd 's dxy 10' specifies the s/dxy/fxyz components\n"
                )
            ret += spd_dict[l]

    return list(set(ret))


def WeightFromPro(infile='PROCAR', whichAtom=None, spd=None, lsorbit=False, spin=None):
    """
    Contribution of selected atoms to the each KS orbital
    """

    assert os.path.isfile(infile), '%s cannot be found!' % infile
    FileContents = [line for line in open(infile) if line.strip()]

    # when the band number is too large, there will be no space between ";" and
    # the actual band number. A bug found by Homlee Guo.
    # Here, #kpts, #bands and #ions are all integers
    nkpts, nbands, nions = [int(xx) for xx in re.sub(
        '[^0-9]', ' ', FileContents[1]).split()]

    if spd:
        Weights = np.asarray([line.split()[1:-1] for line in FileContents
                              if not re.search('[a-zA-Z]', line)], dtype=float)
        Weights = np.sum(Weights[:, spd], axis=1)
    else:
        Weights = np.asarray([line.split()[-1] for line in FileContents
                              if not re.search('[a-zA-Z]', line)], dtype=float)

    nspin = Weights.shape[0] // (nkpts * nbands * nions)
    nspin //= 4 if lsorbit else 1

    if lsorbit:
        Weights.resize(nspin, nkpts, nbands, 4, nions)
        wid = [None, 'x', 'y', 'z'].index(spin)
        Weights = Weights[:, :, :, wid, :]
    else:
        Weights.resize(nspin, nkpts, nbands, nions)

    if whichAtom is None:
        return np.sum(Weights, axis=-1)
    else:
        # whichAtom = [xx - 1 for xx in whichAtom]
        whichAtom = [xx for xx in whichAtom]
        return np.sum(Weights[:, :, :, whichAtom], axis=-1)


############################################################


def get_bandInfo(inFile='OUTCAR'):
    """
    extract band energies from OUTCAR
    """

    outcar = [line for line in open(inFile) if line.strip()]

    for ii, line in enumerate(outcar):
        if 'NKPTS =' in line:
            nkpts = int(line.split()[3])
            nband = int(line.split()[-1])

        if 'ISPIN  =' in line:
            ispin = int(line.split()[2])

        if "k-points in reciprocal lattice and weights" in line:
            Lvkpts = ii + 1

        if 'reciprocal lattice vectors' in line:
            ibasis = ii + 1

        if 'E-fermi' in line:
            Efermi = float(line.split()[2])
            LineEfermi = ii + 1
            # break

    # basis vector of reciprocal lattice
    # B = np.array([line.split()[3:] for line in outcar[ibasis:ibasis+3]],

    # When the supercell is too large, spaces are missing between real space
    # lattice constants. A bug found out by Wei Xie (weixie4@gmail.com).
    B = np.array([line.split()[-3:] for line in outcar[ibasis:ibasis + 3]],
                 dtype=float)
    # k-points vectors and weights
    tmp = np.array([line.split() for line in outcar[Lvkpts:Lvkpts + nkpts]],
                   dtype=float)
    vkpts = tmp[:, :3]
    wkpts = tmp[:, -1]

    # for ispin = 2, there are two extra lines "spin component..."
    N = (nband + 2) * nkpts * ispin + (ispin - 1) * 2
    bands = []
    # vkpts = []
    for line in outcar[LineEfermi:LineEfermi + N]:
        if 'spin component' in line or 'band No.' in line:
            continue
        if 'k-point' in line:
            # vkpts += [line.split()[3:]]
            continue
        bands.append(float(line.split()[1]))

    bands = np.array(bands, dtype=float).reshape((ispin, nkpts, nband))

    if os.path.isfile('KPOINTS'):
        kp = open('KPOINTS').readlines()

    if os.path.isfile('KPOINTS') and kp[2][0].upper() == 'L':
        Nk_in_seg = int(kp[1].split()[0])
        Nseg = nkpts // Nk_in_seg
        vkpt_diff = np.zeros_like(vkpts, dtype=float)

        for ii in range(Nseg):
            start = ii * Nk_in_seg
            end = (ii + 1) * Nk_in_seg
            vkpt_diff[start:end, :] = vkpts[start:end, :] - vkpts[start, :]

        kpt_path = np.linalg.norm(np.dot(vkpt_diff, B), axis=1)
        # kpt_path = np.sqrt(np.sum(np.dot(vkpt_diff, B)**2, axis=1))
        for ii in range(1, Nseg):
            start = ii * Nk_in_seg
            end = (ii + 1) * Nk_in_seg
            kpt_path[start:end] += kpt_path[start - 1]

        # kpt_path /= kpt_path[-1]
        kpt_bounds = np.concatenate((kpt_path[0::Nk_in_seg], [kpt_path[-1], ]))
    else:
        # get band path
        vkpt_diff = np.diff(vkpts, axis=0)
        kpt_path = np.zeros(nkpts, dtype=float)
        kpt_path[1:] = np.cumsum(np.linalg.norm(np.dot(vkpt_diff, B), axis=1))
        # kpt_path /= kpt_path[-1]

        # get boundaries of band path
        xx = np.diff(kpt_path)
        kpt_bounds = np.concatenate(
            ([0.0, ], kpt_path[1:][np.isclose(xx, 0.0)], [kpt_path[-1], ]))

    return kpt_path, bands, Efermi, kpt_bounds


############################################################


def bandplot(kpath, bands, efermi, kpt_bounds, opts, whts=None):
    '''
    Use matplotlib to plot band structure
    '''

    width, height = opts.figsize
    ymin, ymax = opts.ylim
    dpi = opts.dpi

    fig = plt.figure()
    fig.set_size_inches(width, height)
    ax = plt.subplot(111)

    nspin, nkpts, nbands = bands.shape

    clrs = ['r', 'b']

    if opts.occLC and (whts is not None):
        from matplotlib.collections import LineCollection
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        LW = opts.occLC_lw
        DELTA = 0.3
        EnergyWeight = whts[0]
        norm = mpl.colors.Normalize(
            vmin=opts.occLC_cbar_vmin if opts.occLC_cbar_vmin else EnergyWeight.min(),
            vmax=opts.occLC_cbar_vmax if opts.occLC_cbar_vmax else EnergyWeight.max(),
        )
        # norm = mpl.colors.Normalize(0, 1)
        # create a ScalarMappable and initialize a data structure
        s_m = mpl.cm.ScalarMappable(cmap=opts.occLC_cmap, norm=norm)
        s_m.set_array([EnergyWeight])

        for Ispin in range(nspin):
            for jj in range(nbands):
                x = kpath
                y = bands[Ispin, :, jj]
                z = EnergyWeight[Ispin, :, jj]

                ax.plot(x, y,
                        lw=LW + 2 * DELTA,
                        color='gray', zorder=1)

                points = np.array([x, y]).T.reshape(-1, 1, 2)
                segments = np.concatenate([points[:-1], points[1:]], axis=1)
                lc = LineCollection(segments,
                                    # cmap=opts.occLC_cmap, # alpha=0.7,
                                    colors=[s_m.to_rgba(ww)
                                            for ww in (z[1:] + z[:-1]) / 2.]
                                    # norm=plt.Normalize(0, 1)
                                    )
                # lc.set_array((z[1:] + z[:-1]) / 2)
                lc.set_linewidth(LW)
                ax.add_collection(lc)

        divider = make_axes_locatable(ax)
        ax_cbar = divider.append_axes(opts.occLC_cbar_pos.lower(),
                                      size=opts.occLC_cbar_size, pad=opts.occLC_cbar_pad)

        if opts.occLC_cbar_pos.lower() == 'top' or opts.occLC_cbar_pos.lower() == 'bottom':
            ori = 'horizontal'
        else:
            ori = 'vertical'
        cbar = plt.colorbar(s_m, cax=ax_cbar,
                            # ticks=[0.0, 1.0],
                            orientation=ori)
        if opts.occLC_cbar_ticks:
            cbar.set_ticks([
                float(x) for x in
                opts.occLC_cbar_ticks.split()
            ])
            if opts.occLC_cbar_ticklabels:
                cbar.set_ticklabels(opts.occLC_cbar_ticklabels.split())

        if ori == 'horizontal':
            cbar.ax.xaxis.set_ticks_position('top')
        else:
            cbar.ax.yaxis.set_ticks_position('right')


    else:
        for Ispin in range(nspin):
            for Iband in range(nbands):
                # if Iband == 0 else line.get_color()
                lc = opts.linecolors[Ispin]
                line, = ax.plot(kpath, bands[Ispin, :, Iband], lw=opts.linewidth, zorder=0,
                                alpha=0.8,
                                color=lc,
                                )
                if whts is not None:
                    for ii in range(len(opts.occ)):
                        ax.scatter(kpath, bands[Ispin, :, Iband],
                                   color=opts.occMarkerColor[ii],
                                   s=whts[ii][Ispin, :, Iband] *
                                     opts.occMarkerSize[ii],
                                   marker=opts.occMarker[ii], zorder=1, lw=0.0,
                                   alpha=0.5)

    for bd in kpt_bounds:
        ax.axvline(x=bd, ls='-', color='k', lw=0.5, alpha=0.5)

    # add extra horizontal/vertical lines
    for xx in opts.hlines:
        ax.axhline(y=xx, ls=':', color='k', lw=0.5, alpha=0.5)
    for yy in opts.vlines:
        ax.axhline(x=yy, ls=':', color='k', lw=0.5, alpha=0.5)

    if opts.xlabel:
        ax.set_xlabel(opts.xlabel)
    if opts.ylabel:
        ax.set_ylabel(opts.ylabel)
    else:
        ax.set_ylabel('Energy (eV)',  # fontsize='small',
                      labelpad=5)
    ax.set_ylim(ymin, ymax)
    ax.set_xlim(kpath.min(), kpath.max())

    ax.set_xticks(kpt_bounds)
    if opts.kpts:
        kname = [x.upper() for x in opts.kpts]
        for ii in range(len(kname)):
            if kname[ii] == 'G':
                kname[ii] = r'$\mathrm{\mathsf{\Gamma}}$'
            else:
                kname[ii] = r'$\mathrm{\mathsf{%s}}$' % kname[ii]
        ax.set_xticklabels(kname)
    else:
        ax.set_xticklabels([])

    ax.yaxis.set_minor_locator(AutoMinorLocator(2))

    plt.tight_layout(pad=0.20)
    plt.savefig(opts.bandimage, dpi=opts.dpi)


############################################################


def saveband_dat(kpath, bands, opts, whts=None):
    '''
    save band info to txt files
    '''
    prefix = 'pyband'
    spinSuffix = ['up', 'do']
    nspin, nkpts, nbands = bands.shape

    if nspin == 1:
        if opts.gnuplot:
            with open(prefix + '.dat', 'w') as out:
                # line = "kpts      energy      [projection weight]"
                for jj in range(nbands):
                    line = ''
                    for ii in range(nkpts):
                        line = ''
                        line += '%8.4f ' % kpath[ii]
                        line += '%10.4f' % bands[0, ii, jj]
                        line += '\n'
                        out.write(line)
                    line = '\n'
                    out.write(line)
                line = '\n'
                out.write(line)
                if opts.occ:
                    for kk in range(len(opts.occ)):
                        for jj in range(nbands):
                            line = ''
                            for ii in range(nkpts):
                                line = ''
                                line += '%8.4f ' % kpath[ii]
                                line += '%10.4f ' % bands[0, ii, jj]
                                line += '%10.4f' % whts[kk][0, ii, jj]
                                line += '\n'
                                out.write(line)
                            line = '\n'
                            out.write(line)
                        line = '\n'
                        out.write(line)
        else:
            header = "set xran [{}:{}]\n".format(kpath.min(), kpath.max())
            header += "plot for [ii=2:{}] 'pyband.dat' u 1:ii with line lc rgb '#000' t ''\n".format(
                nbands + 1)
            np.savetxt(prefix + '.dat', np.c_[kpath, bands[0]], fmt='%10.4f',
                       header=header)
            if whts:
                for i in range(len(whts)):
                    np.savetxt(prefix + '_weights_component_{}.dat'.format(i + 1),
                               np.c_[kpath, whts[i][0]], fmt='%10.4f', header=header)

    else:
        if opts.gnuplot:
            for Ispin in range(nspin):
                filename = prefix + '_' + spinSuffix[Ispin] + '.dat'
                with open(filename, 'w') as out:
                    # line = "kpts      energy      [projection weight]"
                    for jj in range(nbands):
                        line = ''
                        for ii in range(nkpts):
                            line = ''
                            line += '%8.4f ' % kpath[ii]
                            line += '%10.4f' % bands[Ispin, ii, jj]
                            line += '\n'
                            out.write(line)
                        line = '\n'
                        out.write(line)
                    line = '\n'
                    out.write(line)
                    if opts.occ:
                        for kk in range(len(opts.occ)):
                            for jj in range(nbands):
                                line = ''
                                for ii in range(nkpts):
                                    line = ''
                                    line += '%8.4f ' % kpath[ii]
                                    line += '%10.4f' % bands[Ispin, ii, jj]
                                    line += '%10.4f' % whts[kk][Ispin, ii, jj]
                                    line += '\n'
                                    out.write(line)
                                line = '\n'
                                out.write(line)
                            line = '\n'
                            out.write(line)
        else:
            for Ispin in range(nspin):
                filename = prefix + '_' + spinSuffix[Ispin] + '.dat'
                header = "set xran [{}:{}]\n".format(kpath.min(), kpath.max())
                header += "plot for [ii=2:{}] '{}' u 1:ii with line lc rgb '#000' t ''\n".format(
                    nbands + 1, filename)
                np.savetxt(filename, np.c_[kpath, bands[Ispin]], fmt='%10.4f',
                           header=header)
                if whts:
                    for i in range(len(whts)):
                        filename = prefix + '_' + spinSuffix[Ispin] + '_weights_component_{}'.format(i + 1) + '.dat'
                        np.savetxt(filename, np.c_[kpath, whts[i][Ispin]], fmt='%10.4f',
                                   header=header)



from argparse import ArgumentParser

import requests as rq
import socket
import textwrap
from re import match
import json


class CLICommand:
    """苏大校园网注销脚本
    """

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        pass

    @staticmethod
    def run(args, parser):
        print("--苏大校园网注销脚本--")
        logout()


def logout():
    with socket.socket(socket.AF_INET, socket.SOCK_DGRAM) as s:
        s.connect(('8.8.8.8', 80))
        ip = s.getsockname()[0]
    url = "http://10.9.1.3:801/eportal/?c=Portal&a=logout&callback=dr1004&login_method=1" \
          "&user_account=drcom&user_password=123&ac_logout=1&register_mode=1" \
          f"&wlan_user_ip={ip}&wlan_user_ipv6=&wlan_vlan_id=1" \
          "&wlan_user_mac=000000000000&wlan_ac_ip=&wlan_ac_name=&jsVersion=3.3.3&v=9364"
    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.101 Safari/537.36"
    }
    response = rq.get(url, headers=headers)
    print(parse(response.text))
    response.close()


def parse(s):
    rt = match(r".*?\((\{.*?})\)", s)
    data = rt.group(1)
    data = json.loads(data)
    return textwrap.dedent(f"""
    ------------------------
    msg: {data["msg"]}
    code: {data["result"]}
    ------------------------
    """).strip()



import json
import os
import subprocess
import uuid
from pathlib import Path
import yaml

from htscf.utils.logging import Logger
from htscf.utils.tools import del_nested_attr


class Workflow:
    def __init__(self, workflow_definition):
        self.workflow = workflow_definition
        self.flow_id = f"{self.workflow['name']}_{uuid.uuid4()}"
        self.flow_path = Path(self.flow_id)
        self.flow_path.mkdir(exist_ok=True, parents=True)
        self.logger = Logger()

    def set_global_env(self):
        for var, value in self.workflow.get("env", {}).items():
            os.environ[var] = str(value)

    def run(self):
        self.set_global_env()
        prev_step_info = {}
        for step in self.workflow['steps']:
            for step_name, step_details in step.items():
                current_step_info = self.prepare_step(step_name, step_details)
                if not self.run_step(step_name, current_step_info, prev_step_info):
                    return  # Stop the flow if any step fails
                prev_step_info = current_step_info

    def prepare_step(self, step_name, step_details):
        step_id = f"{step_name}_{uuid.uuid4()}"
        step_path = self.flow_path / step_id
        step_path.mkdir(exist_ok=True, parents=True)
        return {
            "flow_id": self.flow_id,
            "flow_path": self.flow_path.absolute().as_posix(),
            "step_id": step_id,
            "step_path": step_path.absolute().as_posix(),
            "step_details": step_details
        }

    def run_step(self, step_name, current_step_info, prev_step_info):
        self.logger.info(f"Running step: {step_name}")
        self.set_env(current_step_info)
        self.create_step_files(current_step_info)
        try:
            self.exec_script(current_step_info, prev_step_info)
        except Exception as e:
            self.logger.error({
                "content": f"Step '{step_name}' interrupted: {e}",
                "current_step_info": current_step_info
            })
            return False
        return True

    @staticmethod
    def set_env(current_step_info):
        for var, value in current_step_info["step_details"].get("env", {}).items():
            os.environ[var] = str(value)

    @staticmethod
    def create_step_files(current_step_info):
        step_path = Path(current_step_info["step_path"])
        for fileName, content in current_step_info["step_details"].get("files", {}).items():
            (step_path / fileName).write_text(content, encoding="utf-8")

        run = current_step_info["step_details"].get("run", "")
        (step_path / "run").write_text(run, encoding="utf-8")

    def exec_script(self, current_step_info, prev_step_info):
        context = self.get_context_before_run(current_step_info, prev_step_info)
        with open(context["stdoutPath"], "a+", encoding="utf-8") as stdout, \
                open(context["stderrPath"], "a+", encoding="utf-8") as stderr:
            env = os.environ.copy()
            env.update(context["info"])
            subprocess.Popen([context['shell'], context["scriptPath"]],
                             stdout=stdout, stderr=stderr, env=env, cwd=context["step_path"]).communicate()
            if stderr.tell():
                stdout.seek(0)
                stderr.seek(0)
                raise Exception(json.dumps({"stdout": stdout.read(), "stderr": stderr.read()}))

    @staticmethod
    def get_context_before_run(current_step_info, prev_step_info):
        del_nested_attr(current_step_info, "step_details.files")
        del_nested_attr(prev_step_info, "step_details.files")
        step_path = Path(current_step_info["step_path"])
        return {
            "scriptPath": step_path / "run",
            "stdoutPath": step_path / "stdout",
            "stderrPath": step_path / "stderr",
            "step_path": step_path,
            "shell": current_step_info["step_details"].get("shell", ""),
            "info": {
                "current_step_info": json.dumps(current_step_info),
                "prev_step_info": json.dumps(prev_step_info)
            }
        }



"""
This package is designed for logging purposes and enables the seamless uploading of log entries to a MongoDB database.
"""
from functools import partial
from typing import Protocol, Any

from htscf.db.mongo import log_message
from htscf.utils.constants import LOG_DB_CONSTANTS, LOG_LEVEL_CONSTANTS


class ILogger(Protocol):
    def log(self, message: Any, level: int = LOG_LEVEL_CONSTANTS.LOG_LEVEL_INFO) -> None:
        ...

    def __call__(self, message: Any, level: int = LOG_LEVEL_CONSTANTS.LOG_LEVEL_INFO) -> None:
        ...


class __Logger:
    def __init__(self,
                 log_db_name: str,
                 log_collection_name: str,
                 db_host: str,
                 db_port: int):
        self.log_db_name = log_db_name or LOG_DB_CONSTANTS.LOG_DB_NAME
        self.log_collection_name = log_collection_name or LOG_DB_CONSTANTS.LOG_COLLECTION_NAME
        self.db_host = db_host or LOG_DB_CONSTANTS.LOG_DB_HOST
        self.db_port = db_port or LOG_DB_CONSTANTS.LOG_DB_PORT
        self.logger_func = partial(log_message, log_db_name=self.log_db_name, log_collection_name=self.log_collection_name, db_host=self.db_host, db_port=self.db_port)

    def log(self, message: Any, level: int = LOG_LEVEL_CONSTANTS.LOG_LEVEL_INFO) -> None:
        self.logger_func(message=message, level=level)

    def __call__(self, message: Any, level: int = LOG_LEVEL_CONSTANTS.LOG_LEVEL_INFO) -> None:
        self.log(message, level)


def _create_logger(log_db_name: str = None, log_collection_name: str = None, db_host: str = None, db_port: int = None) -> ILogger:
    return __Logger(log_db_name, log_collection_name, db_host, db_port)


class Logger:
    def __init__(self, log_db_name: str = None, log_collection_name: str = None, db_host: str = None, db_port: int = None):
        self._logger = _create_logger(log_db_name, log_collection_name, db_host, db_port)

    def emergency(self, message):
        self._logger(message, LOG_LEVEL_CONSTANTS.LOG_LEVEL_EMERGENCY)

    def alert(self, message):
        self._logger(message, LOG_LEVEL_CONSTANTS.LOG_LEVEL_ALERT)

    def critical(self, message):
        self._logger(message, LOG_LEVEL_CONSTANTS.LOG_LEVEL_CRITICAL)

    def error(self, message):
        self._logger(message, LOG_LEVEL_CONSTANTS.LOG_LEVEL_ERROR)

    def warning(self, message):
        self._logger(message, LOG_LEVEL_CONSTANTS.LOG_LEVEL_WARNING)

    def notice(self, message):
        self._logger(message, LOG_LEVEL_CONSTANTS.LOG_LEVEL_NOTICE)

    def info(self, message):
        self._logger(message, LOG_LEVEL_CONSTANTS.LOG_LEVEL_INFO)

    def debug(self, message):
        self._logger(message, LOG_LEVEL_CONSTANTS.LOG_LEVEL_DEBUG)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'UI_YacifEditer.ui'
##
## Created by: Qt User Interface Compiler version 6.2.4
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide2.QtWidgets import (QAbstractItemView, QApplication, QComboBox, QGridLayout,
    QLabel, QLineEdit, QListView, QMainWindow,
    QMenu, QMenuBar, QPushButton, QSizePolicy,
    QStatusBar, QVBoxLayout, QWidget)

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(657, 393)
        self.actionOpenFile = QAction(MainWindow)
        self.actionOpenFile.setObjectName(u"actionOpenFile")
        self.actionExit = QAction(MainWindow)
        self.actionExit.setObjectName(u"actionExit")
        self.actionAbout = QAction(MainWindow)
        self.actionAbout.setObjectName(u"actionAbout")
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.gridLayout_2 = QGridLayout()
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.filePath = QLineEdit(self.centralwidget)
        self.filePath.setObjectName(u"filePath")

        self.gridLayout_2.addWidget(self.filePath, 0, 1, 1, 1)

        self.label_3 = QLabel(self.centralwidget)
        self.label_3.setObjectName(u"label_3")

        self.gridLayout_2.addWidget(self.label_3, 2, 0, 1, 1)

        self.savePath = QLineEdit(self.centralwidget)
        self.savePath.setObjectName(u"savePath")

        self.gridLayout_2.addWidget(self.savePath, 2, 1, 1, 1)

        self.nextFile = QPushButton(self.centralwidget)
        self.nextFile.setObjectName(u"nextFile")

        self.gridLayout_2.addWidget(self.nextFile, 3, 4, 1, 1)

        self.label = QLabel(self.centralwidget)
        self.label.setObjectName(u"label")

        self.gridLayout_2.addWidget(self.label, 0, 0, 1, 1)

        self.label_2 = QLabel(self.centralwidget)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout_2.addWidget(self.label_2, 3, 0, 1, 1)

        self.fileName = QComboBox(self.centralwidget)
        self.fileName.setObjectName(u"fileName")

        self.gridLayout_2.addWidget(self.fileName, 3, 1, 1, 1)

        self.showYacif = QPushButton(self.centralwidget)
        self.showYacif.setObjectName(u"showYacif")

        self.gridLayout_2.addWidget(self.showYacif, 3, 3, 1, 1)

        self.openSavePath = QPushButton(self.centralwidget)
        self.openSavePath.setObjectName(u"openSavePath")

        self.gridLayout_2.addWidget(self.openSavePath, 2, 3, 1, 1)

        self.openFilePath = QPushButton(self.centralwidget)
        self.openFilePath.setObjectName(u"openFilePath")

        self.gridLayout_2.addWidget(self.openFilePath, 0, 3, 1, 1)


        self.verticalLayout.addLayout(self.gridLayout_2)

        self.listView = QListView(self.centralwidget)
        self.listView.setObjectName(u"listView")
        self.listView.setEditTriggers(QAbstractItemView.NoEditTriggers)

        self.verticalLayout.addWidget(self.listView)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 657, 26))
        self.menu = QMenu(self.menubar)
        self.menu.setObjectName(u"menu")
        self.menu_2 = QMenu(self.menubar)
        self.menu_2.setObjectName(u"menu_2")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)

        self.menubar.addAction(self.menu.menuAction())
        self.menubar.addAction(self.menu_2.menuAction())
        self.menu.addAction(self.actionOpenFile)
        self.menu.addAction(self.actionExit)
        self.menu_2.addAction(self.actionAbout)

        self.retranslateUi(MainWindow)
        self.actionOpenFile.triggered.connect(MainWindow.openFileSlot)
        self.openFilePath.clicked.connect(MainWindow.openFileSlot)
        self.openSavePath.clicked.connect(MainWindow.openFileSlot)
        self.filePath.textChanged.connect(MainWindow.loadYacifSlot)
        self.fileName.currentIndexChanged.connect(MainWindow.openYacifSlot)
        self.showYacif.clicked.connect(MainWindow.openYacifSlot)
        self.nextFile.clicked.connect(MainWindow.nextYacifSlot)
        self.listView.doubleClicked.connect(MainWindow.deleteRowSlot)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"Yacif\u7f16\u8f91\u5668", None))
        self.actionOpenFile.setText(QCoreApplication.translate("MainWindow", u"Open", None))
        self.actionExit.setText(QCoreApplication.translate("MainWindow", u"Exit", None))
        self.actionAbout.setText(QCoreApplication.translate("MainWindow", u"About", None))
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Save Path\uff1a", None))
        self.nextFile.setText(QCoreApplication.translate("MainWindow", u"Next", None))
        self.label.setText(QCoreApplication.translate("MainWindow", u"Path\uff1a", None))
        self.label_2.setText(QCoreApplication.translate("MainWindow", u"Yacif:", None))
        self.showYacif.setText(QCoreApplication.translate("MainWindow", u"Show", None))
        self.openSavePath.setText(QCoreApplication.translate("MainWindow", u"Open", None))
        self.openFilePath.setText(QCoreApplication.translate("MainWindow", u"Open", None))
        self.menu.setTitle(QCoreApplication.translate("MainWindow", u"\u6587\u4ef6", None))
        self.menu_2.setTitle(QCoreApplication.translate("MainWindow", u"\u5e2e\u52a9", None))
    # retranslateUi




#!/usr/bin/env python

from __future__ import print_function

import matplotlib as mpl
from ase.io import read

mpl.use('agg')
mpl.rcParams['axes.unicode_minus'] = False

from matplotlib.patches import Polygon
import matplotlib.colors as mcolors
import os
import re
from argparse import ArgumentParser
from optparse import OptionParser
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib.ticker import AutoMinorLocator
import numpy as np
from time import time

############################################################
__version__ = "1.0"


############################################################


def parseList(string):
    def parseRange(rng):
        # print(rng)
        m = re.match(r'(\d+)(?:[-:](\d+))?(?:[-:](\d+))?$', rng)
        if not m:
            raise ValueError(
                """
The band index should be assigned with combination of the following ways:
-> 10, a single band with index 10
-> 20:30, or '20-30', a continuous range from 20 to 30, 30 included
-> 30:50:2, or '30-50:2', a continues range from 30 to 50, with step size 2
-> '1 2 3', all the patterns above in one string separated by spaces.
For example: '1 4:6 8:16:2' will be converted to '1 4 5 6 8 10 12 14 16'
"""
            )
        ii = m.group(1)
        jj = m.group(2) or ii
        ss = m.group(3) or 1
        return [x - 1 for x in range(int(ii), int(jj) + 1, int(ss))]

    ret = []
    for rng in string.split():
        ret += parseRange(rng)
    return list(set(ret))


def getElemIdx(poscar='POSCAR'):
    '''
    Get the atom indices corresponding to elements in POSCAR
    - poscar: POSCAR file name, default is 'POSCAR'

    returns a dictionary with keys are symbols, vals are list of
    corresponding indices, starting from 0

    e.g. {  'Mo': [0, 1],
            'S': [2, 3, 4, 5],
            'Se': [6, 7],
            'W': [8]}

    Ionizing
    '''
    pos = read(poscar)
    symb_num = {}
    for idx, elem in enumerate(pos.get_chemical_symbols()):
        if elem not in symb_num:
            symb_num[elem] = [idx]
        else:
            symb_num[elem].append(idx)

    return symb_num


def parseSpdProjection(spd):
    '''
    Parse spdProjections string.  str -> [int]

    # Ionizing
    '''
    spd_dict = {
        's': [0],
        'p': [1, 2, 3],
        'd': [4, 5, 6, 7, 8],
        'f': [9, 10, 11, 12, 13, 14, 15],
        'py': [1],
        'pz': [2],
        'px': [3],
        'dxy': [4],
        'dyz': [5],
        'dz2': [6],
        'dxz': [7],
        'dx2': [8],
        "fy(3x2-y2)": [9],
        "fxyz  ": [10],
        "fyz2  ": [11],
        "fz3   ": [12],
        "fxz2  ": [13],
        "fz(x2-y2)": [14],
        "fx(x2-3y2) ": [15],
    }

    ret = []
    for l in spd.split():
        try:
            assert int(l) <= 15, "Maximum spd index should be <= 15."
            ret += [int(l)]
        except:
            if l.lower() not in spd_dict:
                raise ValueError(
                    "Spd-projected wavefunction character of each KS orbital.\n"
                    "    s orbital: 0\n"
                    "    py, pz, px orbital: 1 2 3\n"
                    "    dxy, dyz, dz2, dxz, dx2 orbital: 4 5 6 7 8 \n"
                    "    fy(3x2-y2), fxyz, fyz2, fz3, fxz2, fz(x2-y2), fx(x2-3y2) orbital: 9 10 11 12 13 14 15\n"
                    "\nFor example, --spd 's dxy 10' specifies the s/dxy/fxyz components\n"
                )
            ret += spd_dict[l]

    return list(set(ret))


def WeightFromPro(infile='PROCAR', lsorbit=False):
    """
    Contribution of selected atoms to the each KS orbital
    """

    assert os.path.isfile(infile), '%s cannot be found!' % infile
    FileContents = [line for line in open(infile) if line.strip()]

    # when the band number is too large, there will be no space between ";" and
    # the actual band number. A bug found by Homlee Guo.
    # Here, #kpts, #bands and #ions are all integers
    nkpts, nbands, nions = [int(xx) for xx in re.sub(
        '[^0-9]', ' ', FileContents[1]).split()]

    # Weights = np.asarray([line.split()[-1] for line in FileContents
    #                       if not re.search('[a-zA-Z]', line)], dtype=float)
    Weights = np.asarray([line.split()[1:-1] for line in FileContents
                          if not re.search('[a-zA-Z]', line)], dtype=float)

    kpt_weight = np.asarray(
        [line.split()[-1] for line in FileContents if 'weight' in line], dtype=float)

    energies = np.asarray([line.split()[-4] for line in FileContents
                           if 'occ.' in line], dtype=float)

    nlmax = Weights.shape[-1]
    nspin = Weights.shape[0] // (nkpts * nbands * nions)
    nspin //= 4 if lsorbit else 1

    if lsorbit:
        Weights.resize(nspin, nkpts, nbands, 4, nions, nlmax)
        Weights = Weights[:, :, :, 0, :, :]
    else:
        Weights.resize(nspin, nkpts, nbands, nions, nlmax)

    kpt_weight.resize(nspin, nkpts)
    energies.resize(nspin, nkpts, nbands)

    return energies, kpt_weight, Weights


############################################################


def gradient_fill(x, y, fill_color=None, ax=None, direction=1, **kwargs):
    """
    Plot a line with a linear alpha gradient filled beneath it.

    Parameters
    ----------
    x, y : array-like
        The data values of the line.
    fill_color : a matplotlib color specifier (string, tuple) or None
        The color for the fill. If None, the color of the line will be used.
    ax : a matplotlib Axes instance
        The axes to plot on. If None, the current pyplot axes will be used.
    Additional arguments are passed on to matplotlib's ``plot`` function.

    Returns
    -------
    line : a Line2D instance
        The line plotted.
    im : an AxesImage instance
        The transparent gradient clipped to just the area beneath the curve.
    """

    line, = ax.plot(x, y, **kwargs)
    if fill_color is None:
        fill_color = line.get_color()

    # print fill_color
    zorder = line.get_zorder()
    alpha = line.get_alpha()
    alpha = 1.0 if alpha is None else alpha

    z = np.empty((100, 1, 4), dtype=float)
    rgb = mcolors.colorConverter.to_rgb(fill_color)
    z[:, :, :3] = rgb
    if direction == 1:
        z[:, :, -1] = np.linspace(0, alpha, 100)[:, None]
    else:
        z[:, :, -1] = np.linspace(alpha, 0, 100)[:, None]

    xmin, xmax, ymin, ymax = x.min(), x.max(), y.min(), y.max()
    im = ax.imshow(z, aspect='auto', extent=[xmin, xmax, ymin, ymax],
                   origin='lower', zorder=zorder)

    xy = np.column_stack([x, y])
    if direction == 1:
        xy = np.vstack([[xmin, ymin], xy, [xmax, ymin], [xmin, ymin]])
    else:
        xy = np.vstack([[xmin, ymax], xy, [xmax, ymax], [xmin, ymax]])
    clip_path = Polygon(xy, lw=0.0, facecolor='none',
                        edgecolor='none', closed=True)
    ax.add_patch(clip_path)
    im.set_clip_path(clip_path)

    ax.autoscale(True)

    return line, im


############################################################


def lorentz_smearing(x, x0, sigma=0.03):
    '''
    Lorentz smearing of a Delta function.
    '''
    return sigma / np.pi / ((x - x0) ** 2 + sigma ** 2)


def gaussian_smearing(x, x0, sigma=0.05):
    '''
    Gaussian smearing of a Delta function.
    '''
    smear = np.zeros(x.size)
    condition = np.abs(x - x0) < (sigma * 5.)
    smear[condition] = 1. / (np.sqrt(2 * np.pi) * sigma) * \
                       np.exp(-(x[condition] - x0) ** 2 / (2 * sigma ** 2))

    return smear


def gaussian_smearing_org(x, x0, sigma=0.05):
    '''
    Gaussian smearing of a Delta function.
    '''

    return 1. / (np.sqrt(2 * np.pi) * sigma) * np.exp(-(x - x0) ** 2 / (2 * sigma ** 2))


def generateDos(opts):
    '''
    generate dos
    '''

    ens, kptw, whts = WeightFromPro(opts.procar, opts.lsorbit)
    nspin, nkpts, nbands, nions, nlmax = whts.shape

    emin = ens.min()
    emax = ens.max()
    eran = emax - emin
    emin = emin - eran * opts.extra
    emax = emax + eran * opts.extra

    xen = np.linspace(emin, emax, opts.nedos)
    # tDOS = np.empty((opts.nedos, nspin))
    pDOS = []

    # make all the k-points weight equal
    if opts.homoKpts:
        kptw[...] = 1.0

    tdos_smear = np.empty((nspin, nkpts, nbands, opts.nedos))
    for IS in range(nspin):
        sign = 1 if IS == 0 else -1
        for Ik in range(nkpts):
            for Ib in range(nbands):
                x0 = ens[IS, Ik, Ib]
                # tdos_smear[IS, Ik, Ib] = sign * gaussian_smearing(xen, x0, opts.sigma) * kptw[IS, Ik]
                tdos_smear[IS, Ik, Ib] = sign * \
                                         gaussian_smearing_org(xen, x0, opts.sigma) * kptw[IS, Ik]
                # tdos_smear[IS, Ik, Ib] = sign * lorentz_smearing(xen, x0, opts.sigma) * kptw[IS, Ik]
    tDOS = np.sum(tdos_smear, axis=(1, 2)).T

    # tdos_smear = np.empty((nspin, nkpts, nbands, opts.nedos))
    # for IS in range(nspin):
    #     sign = 1 if IS == 0 else -1
    #     for Ik in range(nkpts):
    #         tdos_smear[IS, Ik, :] = sign * gaussian_smearing_org(xen[np.newaxis, ...],
    #                                         ens[IS,Ik,:,np.newaxis], opts.sigma) * kptw[IS, Ik]
    # tDOS = np.sum(tdos_smear, axis=(1, 2)).T

    # tdos_smear = gaussian_smearing_org(xen[np.newaxis,...], ens[...,np.newaxis],
    #         opts.sigma) * kptw[..., np.newaxis, np.newaxis]
    # tDOS = np.sum(tdos_smear, axis=(1, 2)).T
    # if nspin == 2: tDOS[:,1] *= -1

    if len(opts.elem_list) != 0:
        elem_idx = getElemIdx(opts.posfile)
        for elem in opts.elem_list:
            if elem.upper() == 'XX':
                opts.pdosAtom = [
                    ' '.join([str(i + 1) for i in elem_idx[k]]) for k in elem_idx]
                # print(opts.pdosAtom)
                opts.pdosLabel = list(elem_idx.keys())
                break
            opts.pdosAtom.append(" ".join([str(i + 1) for i in elem_idx[elem]]))
            opts.pdosLabel.append(elem)

    if len(opts.pdosAtom) != 0:

        # add a factor for each PDOS
        factor = np.ones(len(opts.pdosAtom))
        for ii in range(min(len(opts.pdosAtom), len(opts.pdosFactor))):
            factor[ii] = opts.pdosFactor[ii]

        # only plot DOS from selected kpoints, default from all kpoints
        selected_kpts_index = [slice(None, None)
                               for ii in range(len(opts.pdosAtom))]
        for ii in range(min(len(opts.pdosAtom), len(opts.pdosKpts))):
            alist = np.array(opts.pdosKpts[ii].split(), dtype=int)
            # index starting from 1
            assert np.max(
                alist) <= nkpts, "Maximum kpoint index must <{:d}".format(nkpts)
            assert np.min(alist) >= 1, "Minimum kpoint index must >=1"

            nlist = [x for x in alist if not x == -1]
            cmark, = np.where(alist == -1)
            for ii in cmark:
                nlist += range(alist[ii + 1], alist[ii + 2] + 1)
            # index starting from 0
            nlist = [x - 1 for x in set(nlist)]

            selected_kpts_index[ii] = nlist

        for ia, atoms in enumerate(opts.pdosAtom):
            # p = np.zeros((opts.nedos, nspin))

            #  alist = np.array(atoms.split(), dtype=int)
            if '0' in atoms.split():
                nlist = range(nions)
            else:
                nlist = parseList(atoms)
                # nlist = [x for x in alist if not x == -1]
                # cmark, = np.where(alist == -1)
                # for ii in cmark:
                #     nlist += range(alist[ii + 1], alist[ii + 2] + 1)
                # nlist = [x - 1 for x in set(nlist)]

            if ia <= len(opts.spdProjections) - 1:
                # spdList = [int(x) for x in opts.spdProjections[ia].split()]
                spdList = parseSpdProjection(opts.spdProjections[ia])  # Ionizing
                pwhts = np.sum(whts[..., spdList], axis=-1)
            else:
                pwhts = np.sum(whts, axis=-1)

            pwhts = np.sum(pwhts[:, :, :, nlist], axis=-1)

            # p = np.sum(pwhts[..., np.newaxis] * tdos_smear, axis=(1, 2)).T
            p = np.sum(pwhts[:, selected_kpts_index[ia], :, np.newaxis] *
                       tdos_smear[:, selected_kpts_index[ia], ...], axis=(1, 2)).T

            for IS in range(nspin):
                sign = 1 if IS == 0 else -1

                p[:, IS] = opts.pdosOffset * ia * sign + p[:, IS] * factor[ia]
                if ia == 0:
                    tDOS[:, IS] += sign * opts.pdosOffset * len(opts.pdosAtom)

            pDOS += [p]

    return xen, tDOS, pDOS


############################################################


def readDOSFromFile(opts):
    '''
    Read DOS info from file.

    the format of the DOS file:
    first line: ISPIN, NEDOS
    second line: labels of the dos
    next lines:
        if ISPIN = 1:
            Energy pDOS1 PDOS2 ... TotalDOS
        else:
            Energy pDOS1_up PDOS2_up ... TotalDOS_up pDOS1_down PDOS2_down ... TotalDOS_down
    '''

    inp = open(opts.dosFromFile).readlines()

    # the dos basic info
    nspin, nedos = [int(x) for x in inp[0].split()[1:]]
    labels = inp[1].split()[1:]
    # data
    DOS = np.array([line.split() for line in inp[2:] if line.strip()],
                   dtype=float)
    NoPdos = (DOS.shape[1] - 1) // nspin - 1

    tDOS = np.empty((nedos, nspin))
    pDOS = []
    xen = DOS[:, 0]
    for ii in range(nspin):
        tDOS[:, ii] = DOS[:, (ii + 1) * (NoPdos + 1)]
    for pp in range(NoPdos):
        tmp = []
        for ii in range(nspin):
            tmp += [DOS[:, (pp + 1) + ii * (NoPdos + 1)]]
        pDOS += [np.array(tmp).T]

    opts.nedos = nedos
    opts.pdosAtom = ['' for x in range(NoPdos)]
    opts.pdosLabel = labels

    return xen, tDOS, pDOS


def saveDOSToFile(opts, xen, tDOS, pDOS):
    '''
    save DOS info to file.

    the format of the DOS file:
    first line: ISPIN, NEDOS
    second line: labels of the dos
    next lines:
        if ISPIN = 1:
            Energy pDOS1 PDOS2 ... TotalDOS
        else:
            Energy pDOS1_up PDOS2_up ... TotalDOS_up pDOS1_down PDOS2_down ... TotalDOS_down
    '''

    nspin = tDOS.shape[1]
    nedos = tDOS.shape[0]
    NoPdos = len(pDOS)

    out = open(opts.dosToFile, 'w')
    out.write('# %5d %8d\n' % (nspin, nedos))
    labels = '# ' + ' '.join(opts.pdosLabel) + '\n'
    out.write(labels)

    for nn in range(nedos):
        line = '%8.4f ' % xen[nn]
        for ii in range(nspin):
            for p in pDOS:
                line += '%8.4f ' % p[nn, ii]
            line += '%8.4f ' % tDOS[nn, ii]
        line += '\n'
        out.write(line)


############################################################


def dosplot(xen, tdos, pdos, opts):
    '''
    Use matplotlib to plot band structure
    '''

    width, height = opts.figsize
    xmin, xmax = opts.xlim
    dpi = opts.dpi

    plt.style.use(opts.mpl_style)
    # DO NOT use unicode minus regardless of the style
    mpl.rcParams['axes.unicode_minus'] = False

    fig = plt.figure()
    fig.set_size_inches(width, height)
    ax = plt.subplot(111)

    LINES = []
    nspin = tdos.shape[1]

    plabels = []
    LWs = []
    LCs = []
    if opts.pdosAtom:
        plabels = ['p_%d' % ii for ii in range(len(opts.pdosAtom))]
        LWs = [0.5 for ii in range(len(opts.pdosAtom))]
        LCs = [None for ii in range(len(opts.pdosAtom))]

        for ii in range(min(len(opts.pdosAtom), len(opts.pdosLabel))):
            plabels[ii] = opts.pdosLabel[ii]
        for ii in range(min(len(opts.pdosAtom), len(opts.linewidth))):
            LWs[ii] = opts.linewidth[ii]
        for ii in range(min(len(opts.pdosAtom), len(opts.linecolors))):
            LCs[ii] = opts.linecolors[ii]

    plabels += ['total']

    xen -= opts.zero
    for ip, p in enumerate(pdos):
        for ii in range(nspin):
            fill_direction = 1 if ii == 0 else -1
            lc = LCs[ip] if ii == 0 else line.get_color()

            if opts.fill:
                line, im = gradient_fill(xen, p[:, ii], ax=ax, lw=LWs[ip],
                                         color=lc,
                                         direction=fill_direction)
            else:
                line, = ax.plot(xen, p[:, ii], lw=LWs[ip], alpha=0.6,
                                color=lc)
            if ii == 0:
                LINES += [line]

    if opts.showtotal:
        for ii in range(nspin):
            fill_direction = 1 if ii == 0 else -1
            lc = 'k' if ii == 0 else line.get_color()

            if opts.fill:
                line, im = gradient_fill(xen, tdos[:, ii], ax=ax,
                                         color=lc,
                                         lw=0.5,
                                         # zorder=-1,
                                         direction=fill_direction,
                                         )
            else:
                line, = ax.plot(xen, tdos[:, ii], color=lc,
                                lw=0.5, alpha=0.6)
            if ii == 0:
                LINES += [line]
    if opts.xlabel:
        ax.set_xlabel(opts.xlabel)
    else:
        ax.set_xlabel('Energy [eV]',  # fontsize='small',
                      labelpad=5)
    if opts.ylabel:
        ax.set_ylabel(opts.ylabel)
    else:
        ax.set_ylabel('DOS [arb. unit]',  # fontsize='small',
                      labelpad=10)
    ax.tick_params(which='both', labelsize='small')

    ax.set_xlim(xmin, xmax)
    if opts.ylim is not None:
        ymin, ymax = opts.ylim
        ax.set_ylim(ymin, ymax)

    # ax.set_yticklabels([])

    ax.xaxis.set_minor_locator(AutoMinorLocator(2))
    ax.yaxis.set_minor_locator(AutoMinorLocator(2))

    opts.pdosLabel = plabels
    ax.legend(LINES, plabels,
              loc=opts.legendloc,
              fontsize='small',
              frameon=True,
              framealpha=0.6)

    plt.tight_layout(pad=0.50)
    plt.savefig(opts.dosimage, dpi=opts.dpi)


############################################################


class CLICommand:
    """DOS绘制
    """

    @staticmethod
    def add_arguments(parser: ArgumentParser):
        add = parser.add_argument

        add("-i", '--input',
            action='store', type=str, dest='procar',
            default='PROCAR',
            help='location of the PROCAR')

        add("-p", '--pdos',
            action='append', type=str, dest='pdosAtom',
            default=[],
            help='specify which atoms to plot the pdos, 0 for all atoms')

        add("-k", '--kpts',
            action='append', type=str, dest='pdosKpts',
            default=[],
            help='specify which k-points to plot the pdos')

        add("--homokpts",
            action='store_true', dest='homoKpts',
            default=False,
            help='Homogeneous k-points weights')

        add('--pdosoffset',
            action='store', type=float, dest='pdosOffset',
            default=0.0,
            help='offset in pdos plot')

        add("-l", '--label',
            action='append', type=str, dest='pdosLabel',
            default=[],
            help='label of the pdos')

        add('--lloc',
            action='store', type=str, dest='legendloc',
            default='upper right',
            help='legend location of dos plot')

        add('--fac',
            action='append', type=float, dest='pdosFactor',
            default=[],
            help='scale factor of the pdos')

        add('-z', '--zero',
            action='store', type=float,
            dest='zero', default=0.0,
            help='energy reference of the band plot')

        add('--sigma',
            action='store', type=float,
            dest='sigma', default=0.05,
            help='smearing addault 0.05')

        add('-n', '--nedos',
            action='store', type=int,
            dest='nedos', default=5000,
            help='number of point in DOS plot')

        add('-o', '--output',
            action='store', type=str, dest='dosimage',
            default='dos.png',
            help='output image name, "dos.png" by default')

        add('-s', '--size', nargs=2,
            action='store', type=float, dest='figsize',
            default=(4.8, 3.0),
            help='figure size of the output plot')

        add('-x', nargs=2,
            action='store', type=float, dest='xlim',
            default=(-6, 6),
            help='x limit of the dos plot')

        add('-y', nargs=2,
            action='store', type=float, dest='ylim',
            default=None,
            help='energy range of the band plot')

        add('-e',
            action='store', type=float, dest='extra',
            default=0.05,
            help='extra energy range of the band plot')

        add('--lw',
            action='append', type=float, dest='linewidth',
            default=[],
            help='linewidth of the band plot')

        add('--lc',
            action='append', type=str, dest='linecolors',
            default=[],
            help='linecolors of the band plot')

        add('--fill',
            action='store_true', dest='fill',
            default=True,
            help='fill under the DOS')

        add('--nofill',
            action='store_false', dest='fill',
            help='no fill under the DOS')

        add('--dpi',
            action='store', type=int, dest='dpi',
            default=360,
            help='resolution of the output image')

        add('--tot',
            action='store_true', dest='showtotal',
            default=True,
            help='show total dos')

        add('--notot',
            action='store_false', dest='showtotal',
            help='not show total dos')

        add('--style',
            action='store', type=str, dest='mpl_style',
            default='default',
            help='plot style of matplotlib. See "plt.style.available" for list of available styles.')

        add('--fromfile',
            action='store', type=str, dest='dosFromFile',
            default=None,
            help='plot the dos contained in the file')

        add('--tofile',
            action='store', type=str, dest='dosToFile',
            default=None,
            help='save DOS to file.')

        add('--spd',
            action='append', type=str, dest='spdProjections',
            default=[],
            help="Spd-projected wavefunction character of each KS orbital.\n"
                 "    s orbital: 0\n"
                 "    py, pz, px orbital: 1 2 3\n"
                 "    dxy, dyz, dz2, dxz, dx2 orbital: 4 5 6 7 8 \n"
                 "    fy(3x2-y2), fxyz, fyz2, fz3, fxz2, fz(x2-y2), fx(x2-3y2) orbital: 9 10 11 12 13 14 15\n"
                 "\nFor example, --spd 's dxy 9\n"
            )

        add('--lsorbit',
            action='store_true', dest='lsorbit',
            help='Spin orbit coupling on, special treament of PROCAR')

        add('-q', '--quiet',
            action='store_true', dest='quiet',
            help='not show the resulting image')

        add('--elem',
            action='append', type=str, dest='elem_list',
            default=[],
            help='display the PDOS of given element(s), if --elem == "XX", show all of them respectively.')  # Ionizing

        add('--poscar',
            action='store', type=str, dest='posfile',
            default='POSCAR',
            help='specify which poscar to read, if "--elem" is not empty')  # Ionizing

        add('--xlabel',
            action='store', type=str,
            dest='xlabel', default=None,
            help='设置x坐标描述')

        add('--ylabel',
            action='store', type=str,
            dest='ylabel', default=None,
            help='设置y坐标描述')

    @staticmethod
    def run(args, parser):
        if args.dosFromFile:
            xen, tdos, pdos = readDOSFromFile(args)
        else:
            t0 = time()
            xen, tdos, pdos = generateDos(args)
            t1 = time()
            print('DOS calc completed! Time Used: %.2f [sec]' % (t1 - t0))

        t0 = time()
        dosplot(xen, tdos, pdos, args)
        t1 = time()
        print('DOS plot completed! Time Used: %.2f [sec]' % (t1 - t0))

        # save dos to file
        if args.dosToFile:
            saveDOSToFile(args, xen, tdos, pdos)



from copy import deepcopy
from io import BytesIO
from multiprocessing import Process
from pathlib import PurePath, Path
from typing import IO
from typing import Union

import yaml
from ase.gui.gui import GUI
from ase.io import read
from ase.io.cif import parse_cif_ase

from htsct.core.buildRibbon import RibbonBuilder
from htsct.utils.tools import md5_hex, parse_cif_ase_string

NameOrFile = Union[str, PurePath, IO]


class ASEGUI(GUI):
    def __init__(self, parent=None, atoms=None, repeat_images=(3, 1, 1), *args, **kwargs):
        super().__init__(*args, **kwargs)
        if isinstance(atoms, (str, PurePath, IO)):
            atoms = read(atoms)
        self.parent = parent
        self.new_atoms(atoms)
        self.images.repeat_images(repeat_images)
        self.set_frame()
        self.toggle_show_bonds(self)

    def show(self):
        if self.parent:
            self.parent.showStructure.setText("Close")
        self.run()

    def close(self):
        self.exit()

    def exit(self, event=None):
        super().exit(event)
        if self.parent:
            self.parent.showStructure.setText("Show")
            self.parent.state["subGUIState"] = False
            del self.parent.ase_gui
            self.parent.ase_gui = None


class GUIProcess(Process):
    def __init__(self, guiConfig, *args, **kwargs):
        super(GUIProcess, self).__init__(*args, **kwargs)
        self.guiConfig = guiConfig
        for k, v in guiConfig.items():
            if "currentIndex" in v.keys():
                d = v["data"][v["currentIndex"]]
            else:
                d = v["data"]
            setattr(self, k, d)
        self.repeatImages = (int(self.repeatA), int(self.repeatB), int(self.repeatC))
        self.filePath = Path(self.rootPath) / self.choiceStructure
        self.connectivity = (float(self.connectivityMin), float(self.connectivityMax))
        self.bottom = int(self.bottom)
        self.top = int(self.top)
        self.widthControlBase = int(self.widthControlBase)
        self.widthControlDelta = int(self.widthControlDelta)
        self.widthControlOrigo = int(self.widthControlOrigo)
        self.milerIndex = (
            list(map(lambda x: int(x), self.vectorA.split(","))),
            list(map(lambda x: int(x), self.vectorB.split(","))),
            list(map(lambda x: int(x), self.vectorC.split(",")))
        )
        self.vacuums = (
            int(self.vacuumA),
            int(self.vacuumB),
            int(self.vacuumC)
        )
        self.ribbonTolerance = float(self.ribbonTolerance)
        self.fixedRibbonWidth = float(self.fixedRibbonWidth)
        self.savedPath = Path(self.savedPath)
        self.saveFilePath = self.savedPath / self.choiceStructure
        self.minRibbonWidth = float(self.minRibbonWidth)
        self.maxRibbonWidth = float(self.maxRibbonWidth)
        self.ifWidthTest = int(self.ifWidthTest)
        self.ifShowTargetStructure = int(self.ifShowTargetStructure)
        self.terminateAtom = self.terminateAtom
        self.data = {}  # 用于储存原始结构数据
        self.loadYamlCif()

    def show_ribbon(self, ribbon):
        ase_gui = ASEGUI(None, ribbon, self.repeatImages)
        ase_gui.show()

    def run(self) -> None:
        for ribbon in self.buildRibbon():
            ribbon_width = self.getRibbonWidth(ribbon)
            print(f"当前纳米带宽度为：{ribbon_width}....")
            if self.fixedRibbonWidth:  # 满足固定宽度ribbon时
                if self.terminateAtom:
                    _ribbon = RibbonBuilder.terminate(ribbon, symbol=self.terminateAtom, side=2,
                                                      connectivity_opts=self.connectivity)
                else:
                    _ribbon = RibbonBuilder.edge_del(ribbon, self.ribbonTolerance)
                fixedWidth = self.fixedRibbonWidth
                ribbon_width = float(f"{ribbon_width:.2f}")
                if ribbon_width > fixedWidth:
                    break
                if fixedWidth <= ribbon_width < fixedWidth + 0.01:
                    if self.savedPath.exists():
                        _ribbon = self.reloadStructure(_ribbon)
                        print(f"达到设置的固定宽度{fixedWidth}!!\n正在保存...")
                        self.show_ribbon(_ribbon)
                        print(f"保存{self.saveStructureToYacif(_ribbon)} 成功!!!")
                    break
                print(f"未达到设置的固定宽度{fixedWidth}...")
                continue
            if self.minRibbonWidth < ribbon_width < self.maxRibbonWidth:  # 满足指定宽度ribbon时
                if self.terminateAtom:
                    _ribbon = RibbonBuilder.terminate(ribbon, symbol=self.terminateAtom, side=2,
                                                      connectivity_opts=self.connectivity)
                else:
                    _ribbon = RibbonBuilder.edge_del(ribbon, self.ribbonTolerance)
                print(f"当前纳米带符合条件：{ribbon_width},正在保存...", end="    ")
                if self.ifShowTargetStructure:
                    self.show_ribbon(_ribbon)
                print(f"保存{self.saveStructureToYacif(_ribbon)} 成功!!!")
                break
            if self.ifWidthTest:  # 当条件不满足时，是否展示结构
                self.show_ribbon(RibbonBuilder.edge_del(ribbon, self.ribbonTolerance))

    def saveStructureToYacif(self, ribbon):
        # 将后缀名统一为".cif",并写入cif文件
        if self.saveFilePath.suffix == ".yacif":
            saveFilePath = self.saveFilePath.__str__().rsplit(".", maxsplit=1)[0] + ".cif"
        else:
            saveFilePath = self.saveFilePath
        ribbon.write(saveFilePath, "cif")
        # 读取cif文件内容
        saveFilePath = Path(saveFilePath)
        cif = saveFilePath.read_text(encoding="utf-8")

        yacifName = f"{self.choiceStructure.rsplit('.', maxsplit=1)[0]}_{md5_hex(saveFilePath)}.yacif"
        yacifPath = self.savedPath / yacifName
        with yacifPath.open(mode="w+", encoding="utf-8") as fd:
            yaml.dump({"guiInfo": self.guiConfig, "cif": cif, "structureInfo": self.data["structureInfo"]}, fd)
        Path(saveFilePath).unlink(missing_ok=True)
        return yacifPath.__str__()

    @staticmethod
    def reloadStructure(atoms):
        with BytesIO() as b1:  # reload structure from cif
            atoms.write(b1, "cif")
            b1.seek(0)
            cif_block = parse_cif_ase(b1)
            return list(cif_block)[0].get_atoms()

    def buildRibbon(self):
        atoms = parse_cif_ase_string(self.data["cif"])
        for i in range(self.bottom, self.top):
            layer = 2 + self.widthControlBase + self.widthControlDelta * i
            ribbon = RibbonBuilder.build(atoms, layer, self.widthControlOrigo, self.ribbonTolerance,
                                         milerIndex=self.milerIndex, vacuums=self.vacuums)
            yield ribbon

    @staticmethod
    def getRibbonWidth(ribbon):
        _ribbon = ribbon.copy()
        _ribbon = RibbonBuilder.edge_del(_ribbon)
        width = RibbonBuilder.get_ribbon_width(_ribbon)
        return width

    def loadYamlCif(self):
        if self.filePath.suffix == ".yacif":

            with open(self.filePath, encoding="utf-8") as fd:
                data = yaml.load(fd, Loader=yaml.FullLoader)
                self.data.update(data)
        else:
            self.data["cif"] = self.filePath.read_text(encoding="utf-8")
            self.data["structureInfo"] = {}



import warnings
from copy import deepcopy
from enum import Enum
from io import BytesIO
from itertools import chain
from pprint import pprint
from typing import Union, List
import ase
import numpy as np
from ase.atoms import Atom
from ase.build import make_supercell, cut
from ase.data import covalent_radii, chemical_symbols
from ase.io.cif import parse_cif_ase
from htscf.utils import vdwbulk

warnings.filterwarnings("ignore")


class RIBBON(Enum):
    """获取边界原子的方向"""
    LEFT = 0
    RIGHT = 1


class F:
    """
    Some helpful Functions for build nano-ribbons.
    """

    @staticmethod
    def get_distance(pos1, pos2):
        """Calculate the distance of two positions."""
        return np.sqrt(np.sum(np.square(pos1 - pos2)))

    @staticmethod
    def find_indexes(atoms, positions):
        """
        Find the atom index from atoms object by their position and return a series of indexes.
        """
        atom_indexes = []
        if isinstance(positions, list):
            for pos in positions:
                for atom in atoms:
                    if (pos == atom.position).all():
                        atom_indexes.append(atom.index)
                        break
            return atom_indexes
        else:
            for atom in atoms:
                if (positions == atom.position).all():
                    return atom.index

    @staticmethod
    def get_covalent_radii(symbol):
        """
        Get covalent_radii by symbol
        :param symbol:symbol,such as 'H'
        """
        if isinstance(symbol, str):
            index = chemical_symbols.index(symbol)
            return covalent_radii[index]
        elif isinstance(symbol, (tuple, set, list)):
            radius = []
            for i in symbol:
                index = chemical_symbols.index(i)
                radius.append(covalent_radii[index])
            return radius

    @staticmethod
    def layered(atoms, axis: int = 0, tol: float = 0.0, pos_type="cartesian") -> List[list]:
        """
        Determine atomic layers based on coordinates
        :param atoms:
        :param pos_type:
        :param axis: Axis of coordinates
        :param tol:Position tolerance
        :return:Layered positions
        """
        if pos_type == "fractional":
            positions = sorted(atoms.get_scaled_positions(), key=lambda x: x[axis])
        else:
            positions = sorted(atoms.positions, key=lambda x: x[axis])

        if axis > len(positions[0]) - 1:
            raise ValueError("The axis greater than positions[0]")
        li = [[positions.pop(0)]]
        while positions:
            element = positions.pop(0)
            last = li[-1][0][axis]
            if abs(element[axis] - last) <= tol:
                li[-1].append(element)
            else:
                li.append([element])
        return li

    @staticmethod
    def find_neighbor_list(atoms):
        neighbor_list = []
        length = len(atoms)
        for i in range(length):
            tmp = set()
            atom1 = atoms[i]
            for j in range(i):
                atom2 = atoms[j]
                symbol1, symbol2 = atom1.symbol, atom2.symbol
                distance = F.get_distance(atom1.position, atom2.position)
                r1, r2 = F.get_covalent_radii(symbol1), F.get_covalent_radii(symbol2)
                if 0 <= distance < r1 + r2:
                    tmp.add(atom1)
                    tmp.add(atom2)
            if tmp:
                neighbor_list.append(tmp)
        return neighbor_list

    @staticmethod
    def extend_vector(a, b, scale=1.0):
        """
        Extend the vector ab and return the point new_b
        """
        return a + (b - a) * scale

    @staticmethod
    def diff(arr1, arr2):
        """
        Find the difference set of two lists, arr1 - arr2.
        """
        if not isinstance(arr1, np.ndarray):
            arr1 = np.array(arr1)
        if not isinstance(arr2, np.ndarray):
            arr2 = np.array(arr2)
        tmp_index = []
        index = []
        for i in range(arr1.shape[0]):
            tmp = arr1[i]
            for j in arr2:
                if (tmp == j).all():
                    tmp_index.append(i)
        for i in range(arr1.shape[0]):
            if i not in tmp_index:
                index.append(i)
        return arr1[index]

    @staticmethod
    def find_conn(arr1, arr2, d=2):
        tmp = []
        for i in arr1:
            for j in arr2:
                distance = F.get_distance(i, j)
                if distance < d:
                    tmp.append([i, j])
        return tmp

    @staticmethod
    def find_indexes_by_scale(atoms: ase.Atoms, positions: Union[np.ndarray, list]) -> Union[int, list]:
        """
        Find the atom index from atoms object by their position and return a series of indexes.
        :param atoms:
        :param positions:
        :return:
        """
        atom_indexes = []
        positions = np.array(positions)
        fracs = atoms.get_scaled_positions()
        for pos in positions:
            for i, pos2 in enumerate(fracs):
                pos = np.round(pos, decimals=2)
                pos2 = np.round(pos2, decimals=2)
                if (pos == pos2).all():
                    atom_indexes.append(i)
        return atom_indexes

    @staticmethod
    def get_edge_pbc_positions(atoms):
        scaled_positions = atoms.get_scaled_positions()  # 获取分数坐标
        sp1 = scaled_positions.copy()  # 分数坐标副本
        sp1 = np.round(sp1, decimals=2)  # 分数坐标保留2位数
        sp1[:, 0] -= 1  # 分数坐标减一
        edge_pbc_positions = []
        for pos in sp1:
            if (pos <= 1).all() and (pos >= 0).all():
                edge_pbc_positions.append(pos)
        if edge_pbc_positions:
            edge_pbc_positions = np.array(edge_pbc_positions)
            orig_positions = []
            epp1 = edge_pbc_positions.copy()
            epp1[:, 0] += 1
            for pos in epp1:
                if (pos <= 1).all() and (pos >= 0).all():
                    orig_positions.append(pos)
            return np.array(orig_positions), edge_pbc_positions
        else:
            return None, None

    @staticmethod
    def layered2atom(atoms, axis: int = 0, tol: float = 0.0):
        """
        Determine atomic layers based on coordinates
        :param atoms:
        :param axis: Axis of coordinates
        :param tol:Position tolerance
        :return:Layered atom list
        """
        atoms_list = sorted((atom for atom in atoms), key=lambda x: x.position[axis])
        layer_list = [[atoms_list.pop(0)]]
        while atoms_list:
            element = atoms_list.pop(0)
            last = layer_list[-1][0].position[axis]
            if abs(element.position[axis] - last) <= tol:
                layer_list[-1].append(element)
            else:
                layer_list.append([element])
        pprint(layer_list)
        return layer_list

    @staticmethod
    def get_edge_atom_indices(atoms, direction: RIBBON = RIBBON.LEFT):
        """获取纳米带x方向指定边界的原子编号"""
        indices = []
        origin_scaled = atoms.get_scaled_positions().copy()
        if direction == RIBBON.LEFT:
            origin_scaled[:, 0] += 1  # 原子整体右移，获取左边界原子
        elif direction == RIBBON.RIGHT:
            origin_scaled[:, 0] -= 1  # 原子整体左移，获取右边界原子
        for i, pos in enumerate(origin_scaled):
            if 0 <= pos[0] <= 1:
                indices.append(i)
        return indices

    @staticmethod
    def add_vacuum(atoms, vacuum=10, axis=2):
        """
        Add vacuum layer to the atoms.
        :param atoms:
        :param vacuum: float. The thickness of the vacuum layer (in Angstrom).
        :param axis: int, 0|1|2. The direction of the vacuum layer
        """
        cell = atoms.get_cell()
        positions = F.layered(atoms, axis=axis)
        width = np.sqrt(np.sum(np.square(positions[0][0] - positions[-1][0])))
        axis_length = np.sqrt(np.dot(cell[axis], cell[axis]))
        orig_vacuum = axis_length - width
        new_vacuum = vacuum - orig_vacuum
        x, y, z = cell
        vectors = [[x, y, z], [y, z, x], [z, x, y]]
        vector0, vector1, vector2 = vectors[axis]
        normal = np.cross(vector1, vector2)
        costheta = np.dot(normal, vector0) / np.sqrt(np.dot(normal, normal) * np.dot(vector0, vector0))
        length = np.sqrt(np.dot(vector0, vector0))
        newlength = length + new_vacuum / costheta
        vector0 *= newlength / length
        atoms.set_cell(cell)
        atoms.center(axis=axis)
        return atoms

    @staticmethod
    def move_edge(atoms):
        """
        用于将正好在右边界的原子移动到左侧，以便于加氢。
        """
        orig, new_pos = F.get_edge_pbc_positions(atoms)
        if isinstance(orig, np.ndarray):
            index = F.find_indexes_by_scale(atoms, orig)
            pbc_atoms = atoms[index]
            del atoms[index]
            scaled = pbc_atoms.get_scaled_positions()
            scaled[:, 0] -= 1
            pbc_atoms.set_scaled_positions(scaled)
            atoms.extend(pbc_atoms)


class RibbonBuilder:
    @staticmethod
    def surface(atoms, milerIndex=((1, 0, 0), (0, 1, 0), (0, 0, 1)), super_cell=None, connectivity=(0.1, 1.3)):
        """
        切割三维材料为二维材料，并改变晶胞形状
        """
        a, b, c = milerIndex
        atoms = atoms.copy()
        if not super_cell:
            super_cell = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        atoms = cut(atoms, a, b, c, nlayers=15)
        atoms.edit()
        cluster = vdwbulk.VdwBulk(atoms, connectivity=connectivity).get_cluster()
        len_cluster = [len(c) for c in cluster]
        index = len_cluster.index(max(len_cluster))
        drop_index = []
        for i, v in enumerate(cluster):
            if i != index:
                drop_index.extend(v)
        del atoms[drop_index]
        atoms = make_supercell(atoms, super_cell)
        with BytesIO() as b1:  # reload structure from cif
            atoms.write(b1, "cif")
            b1.seek(0)
            cif_block = parse_cif_ase(b1)
            atoms = list(cif_block)[0].get_atoms()
        return atoms

    @staticmethod
    def cut_ribbon(atoms, layer, origo, ribbon_tol, axis=1):
        atoms = atoms.copy()
        positions_layered = F.layered(atoms, axis, ribbon_tol)  # Layering 2D material's positions along a specific axis
        saved_layer_numbers = int(
            len(positions_layered) - layer)  # Number of nano-ribbon layers retained after cutting
        saved_positions = positions_layered[origo:-saved_layer_numbers + origo]
        saved_positions = [j for i in saved_positions for j in i]  # 降维
        atom_indexes = F.find_indexes(atoms, saved_positions)
        atoms = atoms[atom_indexes]
        return atoms

    @staticmethod
    def edge_del(atoms, tol=0, axis=1, direction=2):
        positions = F.layered(atoms, axis=axis, tol=tol)
        if direction == 2:
            indexes = F.find_indexes(atoms, positions[0]) + F.find_indexes(atoms, positions[-1])
        elif direction == 1:
            indexes = F.find_indexes(atoms, positions[-1])
        else:
            indexes = F.find_indexes(atoms, positions[0])
        del atoms[indexes]
        return atoms

    @staticmethod
    def get_ribbon_width(atoms):
        positions = F.layered(atoms, axis=1)
        return np.sqrt(np.sum(np.square(positions[0][0] - positions[-1][0])))

    @staticmethod
    def terminate(atoms, symbol="H", side=0, connectivity_opts=(0.6, 1.3), tol=0.):
        _atoms = atoms.copy()
        if side == 2:  # Used to control hydrogenation position
            _atoms = RibbonBuilder.terminate(_atoms, symbol=symbol, side=0, connectivity_opts=connectivity_opts,
                                             tol=tol)
            _atoms.edit()
            _atoms = RibbonBuilder.terminate(_atoms, symbol=symbol, side=1, connectivity_opts=connectivity_opts,
                                             tol=tol)
            _atoms.edit()
            return _atoms
        directions = [[0, 1], [-1, -2]][side]
        a = _atoms.cell[0]
        edge_left, edge_right = _atoms.copy(), _atoms.copy()
        for atom in edge_left:
            atom.position += a
        for atom in edge_right:
            atom.position -= a
        full = chain(_atoms.copy(), edge_left, edge_right)  # Add the periodic atoms on the left and right
        atom_layer = F.layered2atom(full, axis=1, tol=tol)  # Layer atoms
        edge, inner = atom_layer[directions[0]], atom_layer[directions[1]]  # Get boundary and sublayer atoms
        r_new_atom = F.get_covalent_radii(symbol)  # Gets the atomic radius used for passivation
        for atom1 in inner:
            for atom2 in edge:
                # Obtain the covalent bond length between the sublayer and outermost atom
                r1, r2 = F.get_covalent_radii([atom1.symbol, atom2.symbol])
                distance = F.get_distance(atom1.position, atom2.position)
                if (r1 + r2) * connectivity_opts[0] < distance < (r1 + r2) * connectivity_opts[1]:
                    bond_length = r_new_atom + r1
                    terminate_position = F.extend_vector(atom1.position, atom2.position, scale=bond_length / distance)
                    atom = Atom(symbol, position=terminate_position)
                    _atoms.extend(atom)
        del _atoms[[i.index for i in edge]]  # Delete the outermost atoms
        return _atoms

    @staticmethod
    def add_vacuum(atoms, vacuums):
        # _atoms = F.add_vacuum(atoms, vacuums[0], axis=0)
        _atoms = F.add_vacuum(atoms, vacuums[1], axis=1)
        _atoms = F.add_vacuum(_atoms, vacuums[2], axis=2)
        return _atoms

    @staticmethod
    def build(atoms, layer, origo=0, ribbon_tol=0, axis=1, milerIndex=((1, 0, 0), (0, 1, 0), (0, 0, 1)),
              super_cell=((1, 0, 0), (0, 50, 0), (0, 0, 1)), vacuums=(0, 15, 15), connectivity=(0.1, 1.6)):
        _atoms = RibbonBuilder.surface(atoms, milerIndex, super_cell=super_cell, connectivity=connectivity)
        _atoms = RibbonBuilder.cut_ribbon(_atoms, layer, origo, ribbon_tol, axis)
        _atoms = RibbonBuilder.add_vacuum(_atoms, vacuums)
        return _atoms


if __name__ == '__main__':
    # import yaml
    # from htsct.utils.tools import parse_cif_ase_string
    #
    # fp = r"E:\Projects\Work\HighThroughputScreen\htsct\core\1139.C2F2_dff996350a424697ce2eb7eb736a0349.yacif"
    # with open(fp) as fd:
    #     cif = yaml.load(fd, Loader=yaml.FullLoader)["cif"]
    # atoms = parse_cif_ase_string(cif)
    # atoms = RibbonBuilder.build(atoms, layer=13, milerIndex=((1, 1, 0), (1, -1, 0), (0, 0, -1)))
    # RibbonBuilder.terminate(atoms).edit()
    pass



import time
from multiprocessing import Process, Manager
from os import kill
from signal import SIGKILL

from htscf.utils.logging import Logger
from htscf.core.flow import Workflow

logger = Logger()


class FlowHandler:
    def __init__(self, workflow_definition):
        self.manager = Manager()
        self.sharedDict = self.manager.dict()
        self.flow_id = None
        self.sharedDict["workflow_definition"] = workflow_definition

    def run(self):
        p1 = Process(target=FlowHandler.run_flow, args=(self.sharedDict,))
        p2 = Process(target=self._handler, args=(self.sharedDict,))
        p1.start()
        self.sharedDict["p1_pid"] = p1.pid
        p2.start()
        p1.join()
        p2.terminate()  # 确保 _handler 进程也会被终止
        p2.join()
        self.manager.shutdown()  # 显式关闭 Manager

    @staticmethod
    def run_flow(sharedDict):
        workflow_definition = sharedDict["workflow_definition"]
        workflow = Workflow(workflow_definition)
        sharedDict["flow_id"] = workflow.flow_id
        try:
            workflow.run()
        except Exception as e:
            sharedDict["error"] = str(e)
            logger.error({
                "content": f"流程运行错误：{e}"
            })
        finally:
            sharedDict["status"] = "complete"
            logger.info({
                "content": f"流程{workflow.flow_id}运行结束"
            })

    def _handler(self, sharedDict):
        p1_pid = sharedDict["p1_pid"]
        while True:
            if sharedDict.get('status') == 'complete' or sharedDict.get('error'):
                break
            else:
                if not self.handler(sharedDict):
                    kill(p1_pid, SIGKILL)
                    break
                time.sleep(0.5)

    def handler(self, sharedDict):
        return True



from setuptools import setup

with open(r"C:\Users\ZH\Desktop\important\hse_process\htsct-package\requirements.txt") as fd:
    requires = fd.read().split("\n")
setup(
    install_requires=requires
)



ui_config = {
    "rootPath": {
        "className": "QLineEdit",
        "data": "./cifs"
    },
    "choiceStructure": {
        "className": "QComboBox",
        "data": [],
        "currentIndex": -1
    },
    "savedPath": {
        "className": "QLineEdit",
        "data": "./saved"
    },
    "vectorA": {
        "className": "QComboBox",
        "data": [
            "1,0,0",
            "0,1,0",
            "1,1,0",
            "1,1,1",
            "-1,0,0",
            "0,-1,0",
            "1,-1,0",
            "-1,1,0",
            "-1,-1,0",
            "-1,-1,1",
            "-1,-1,-1"
        ],
        "currentIndex": 2
    },
    "vectorB": {
        "className": "QComboBox",
        "data": [
            "1,0,0",
            "0,1,0",
            "1,1,0",
            "1,1,1",
            "-1,0,0",
            "0,-1,0",
            "1,-1,0",
            "-1,1,0",
            "-1,-1,0",
            "-1,-1,1",
            "-1,-1,-1"
        ],
        "currentIndex": 6
    },
    "vectorC": {
        "className": "QComboBox",
        "data": [
            "0,0,1",
            "0,0,-1"
        ],
        "currentIndex": 0
    },
    "widthControlDelta": {
        "className": "QComboBox",
        "data": [
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "10"
        ],
        "currentIndex": 2
    },
    "widthControlOrigo": {
        "className": "QComboBox",
        "data": [
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "10"
        ],
        "currentIndex": 1
    },
    "widthControlBase": {
        "className": "QComboBox",
        "data": [
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            "10"
        ],
        "currentIndex": 1
    },
    "ribbonTolerance": {
        "className": "QLineEdit",
        "data": "0.0"
    },
    "repeatA": {
        "className": "QLineEdit",
        "data": "3"
    },
    "repeatB": {
        "className": "QLineEdit",
        "data": "3"
    },
    "repeatC": {
        "className": "QLineEdit",
        "data": "1"
    },
    "minRibbonWidth": {
        "className": "QLineEdit",
        "data": "20"
    },
    "maxRibbonWidth": {
        "className": "QLineEdit",
        "data": "25"
    },
    "fixedRibbonWidth": {
        "className": "QLineEdit",
        "data": "0"
    },
    "vacuumA": {
        "className": "QLineEdit",
        "data": "0"
    },
    "vacuumB": {
        "className": "QLineEdit",
        "data": "15"
    },
    "vacuumC": {
        "className": "QLineEdit",
        "data": "15"
    },
    "connectivityMin": {
        "className": "QLineEdit",
        "data": "0"
    },
    "connectivityMax": {
        "className": "QLineEdit",
        "data": "1"
    },
    "bottom": {
        "className": "QLineEdit",
        "data": "0"
    },
    "top": {
        "className": "QLineEdit",
        "data": "15"
    },
    "ifWidthTest": {
        "className": "QCheckBox",
        "data": 0
    },
    "ifShowTargetStructure": {
        "className": "QCheckBox",
        "data": 1
    },
    "terminateAtom": {
        "className": "QComboBox",
        "data": [
            "",
            "H",
            "F",
            "Cl",
            "O"
        ],
        "currentIndex": 0
    }
}



import time
from typing import Callable, Union
from pymongo import ASCENDING
from pymongo.collection import Collection
from htscf.utils.logging import Logger
import traceback

logger = Logger()


def send_message(collection, message, priority=1):
    document = {
        "message": message,
        "status": "new",
        "priority": priority,
        "createdAt": time.time()
    }
    collection.insert_one(document)


def listen_for_messages(collection: Collection, callback: Callable, delay: Union[int, float] = 1):
    try:
        while True:
            # 查找一个状态为 "new" 的消息，按创建时间排序
            message = collection.find_one_and_update(
                {"status": "new"},
                {"$set": {"status": "processing"}},
                sort=[("priority", ASCENDING), ("createdAt", ASCENDING)],
                return_document=True  # 确保返回更新后的文档
            )
            if message:
                try:
                    isProcessed = callback(message["message"])
                    if isProcessed:
                        collection.update_one({"_id": message["_id"]}, {"$set": {"status": "processed"}})
                    else:
                        collection.update_one({"_id": message["_id"]}, {"$set": {"status": "new"}})
                except Exception as e:
                    traceback.print_exc()
                    logger.error(f"Error processing message {message['_id']}: {e}")
                    collection.update_one({"_id": message["_id"]}, {"$set": {"status": "error"}})
            else:
                # 没有新消息，短暂休眠
                time.sleep(delay)
    except KeyboardInterrupt:
        logger.info(f"Message Queue interrupt by user")
    except Exception as e:
        logger.error(f"Unexpected error occurred in message queue: {e}")
